#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tray column model for distillation.
"""

__author__ = "Jaffer Ghouse"

import idaes.logger as idaeslog

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.network import Arc, Port
from pyomo.environ import value, Integers, RangeSet, TransformationFactory, \
    Block, Reference, SolverFactory
from pyomo.util.infeasible import log_infeasible_constraints

# Import IDAES cores
from idaes.generic_models.unit_models.distillation import Tray, Condenser, \
    Reboiler
from idaes.generic_models.unit_models.distillation.condenser \
    import CondenserType, TemperatureSpec
from idaes.core import (declare_process_block_class,
                        UnitModelBlockData,
                        useDefault)
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util import get_solver

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("TrayColumn")
class TrayColumnData(UnitModelBlockData):
    """
    Tray Column model.
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not,
**default** = False. Tray column units do not support dynamic behavior."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Indicates whether holdup terms should be constructed or not.
**default** - False. Tray column units do not have defined volume, thus
this must be False."""))
    CONFIG.declare("number_of_trays", ConfigValue(
        default=None,
        domain=In(Integers),
        description="Number of trays in the column",
        doc="""Indicates the number of trays to be constructed.
**default** - None.
**Valid values:** {
Must be integer number.}"""))
    CONFIG.declare("feed_tray_location", ConfigValue(
        default=None,
        domain=In(Integers),
        description="Feed tray location in the column",
        doc="""Indicates the number of trays to be constructed.
**default** - None.
**Valid values:** {
Must be integer number.}"""))
    CONFIG.declare("condenser_type", ConfigValue(
        default=CondenserType.totalCondenser,
        domain=In(CondenserType),
        description="Type of condenser flag",
        doc="""Indicates what type of condenser should be constructed,
**default** - CondenserType.totalCondenser.
**Valid values:** {
**CondenserType.totalCondenser** - Incoming vapor from top tray is condensed
to all liquid,
**CondenserType.partialCondenser** - Incoming vapor from top tray is
partially condensed to a vapor and liquid stream.}"""))
    CONFIG.declare("condenser_temperature_spec", ConfigValue(
        default=None,
        domain=In(TemperatureSpec),
        description="Temperature spec for the condenser",
        doc="""Temperature specification for the condenser,
**default** - TemperatureSpec.none
**Valid values:** {
**TemperatureSpec.none** - No spec is selected,
**TemperatureSpec.atBubblePoint** - Condenser temperature set at
bubble point i.e. total condenser,
**TemperatureSpec.customTemperature** - Condenser temperature at
user specified temperature.}"""))
    CONFIG.declare("has_heat_transfer", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="heat duty to/from tray construction flag.",
        doc="""indicates if there is heat duty to/from the tray,
**default** - False.
**Valid values:** {
**True** - include a heat duty term,
**False** - exclude a heat duty term.}"""))
    CONFIG.declare("has_pressure_change", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="pressure change term construction flag",
        doc="""indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}"""))
    CONFIG.declare("has_liquid_side_draw", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="liquid side draw construction flag.",
        doc="""indicates if there is a liquid side draw from all trays,
**default** - False.
**Valid values:** {
**True** - include a liquid side draw for all trays,
**False** - exclude a liquid side draw for all trays.}"""))
    CONFIG.declare("has_vapor_side_draw", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="vapor side draw construction flag.",
        doc="""indicates if there is a vapor side draw from all trays,
**default** - False.
**Valid values:** {
**True** - include a vapor side draw for all trays,
**False** - exclude a vapor side draw for all trays.}"""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="property package to use for control volume",
        doc="""property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="arguments to use for constructing property packages",
        doc="""a ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))

    def build(self):
        """Build the model.

        Args:
            None
        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(TrayColumnData, self).build()

        # Create set for constructing indexed trays
        if self.config.number_of_trays is not None:
            self.tray_index = RangeSet(1, self.config.number_of_trays)
            self._rectification_index = RangeSet(
                1, self.config.feed_tray_location - 1)
            self._stripping_index = RangeSet(
                self.config.feed_tray_location + 1,
                self.config.number_of_trays)

        else:
            raise ConfigurationError("The config argument number_of_trays "
                                     "needs to specified and cannot be None.")

        # Add trays
        # TODO:
        # 1. Add support for multiple feed tray locations
        # 2. Add support for specifying which trays have side draws
        self.rectification_section = Tray(
            self._rectification_index,
            default={"has_liquid_side_draw":
                     self.config.has_liquid_side_draw,
                     "has_vapor_side_draw":
                     self.config.has_vapor_side_draw,
                     "has_heat_transfer":
                     self.config.has_heat_transfer,
                     "has_pressure_change":
                     self.config.has_pressure_change,
                     "property_package":
                     self.config.property_package,
                     "property_package_args":
                     self.config.property_package_args})

        self.feed_tray = Tray(
            default={"is_feed_tray": True,
                     "has_liquid_side_draw":
                     self.config.has_liquid_side_draw,
                     "has_vapor_side_draw":
                     self.config.has_vapor_side_draw,
                     "has_heat_transfer":
                     self.config.has_heat_transfer,
                     "has_pressure_change":
                     self.config.has_pressure_change,
                     "property_package":
                     self.config.property_package,
                     "property_package_args":
                     self.config.property_package_args})

        self.stripping_section = Tray(
            self._stripping_index,
            default={"has_liquid_side_draw":
                     self.config.has_liquid_side_draw,
                     "has_vapor_side_draw":
                     self.config.has_vapor_side_draw,
                     "has_heat_transfer":
                     self.config.has_heat_transfer,
                     "has_pressure_change":
                     self.config.has_pressure_change,
                     "property_package":
                     self.config.property_package,
                     "property_package_args":
                     self.config.property_package_args})

        # Add condenser
        self.condenser = Condenser(
            default={"condenser_type": self.config.condenser_type,
                     "temperature_spec":
                         self.config.condenser_temperature_spec,
                     "property_package": self.config.property_package,
                     "property_package_args":
                     self.config.property_package_args})

        # Add reboiler
        self.reboiler = Reboiler(
            default={"has_boilup_ratio": True,
                     "has_pressure_change": self.config.has_pressure_change,
                     "property_package": self.config.property_package,
                     "property_package_args":
                     self.config.property_package_args})

        # Add extension to the feed port
        self.feed = Port(extends=self.feed_tray.feed)

        # Add extensions to rectification section

        # Add extensions to stripping section

        # Construct arcs between trays, condenser, and reboiler
        self._make_rectification_arcs()
        self._make_stripping_arcs()
        self._make_feed_arcs()
        self._make_condenser_arcs()
        self._make_reboiler_arcs()

        TransformationFactory("network.expand_arcs").apply_to(self)

    def _make_rectification_arcs(self):
        self._rectification_stream_index = RangeSet(
            1, self.config.feed_tray_location - 2)

        def rule_liq_stream(self, i):
            return {"source": self.rectification_section[i].liq_out,
                    "destination": self.rectification_section[i + 1].liq_in}

        def rule_vap_stream(self, i):
            return {"source": self.rectification_section[i + 1].vap_out,
                    "destination": self.rectification_section[i].vap_in}

        self.rectification_liq_stream = Arc(
            self._rectification_stream_index, rule=rule_liq_stream)
        self.rectification_vap_stream = Arc(
            self._rectification_stream_index, rule=rule_vap_stream)

    def _make_stripping_arcs(self):

        self._stripping_stream_index = RangeSet(
            self.config.feed_tray_location + 1,
            self.config.number_of_trays - 1)

        def rule_liq_stream(self, i):
            return {"source": self.stripping_section[i].liq_out,
                    "destination": self.stripping_section[i + 1].liq_in}

        def rule_vap_stream(self, i):
            return {"source": self.stripping_section[i + 1].vap_out,
                    "destination": self.stripping_section[i].vap_in}

        self.stripping_liq_stream = Arc(
            self._stripping_stream_index, rule=rule_liq_stream)
        self.stripping_vap_stream = Arc(
            self._stripping_stream_index, rule=rule_vap_stream)

    def _make_feed_arcs(self):

        self.feed_liq_in = Arc(
            source=self.rectification_section[
                self.config.feed_tray_location - 1].liq_out,
            destination=self.feed_tray.liq_in
        )

        self.feed_liq_out = Arc(
            source=self.feed_tray.liq_out,
            destination=self.stripping_section[
                self.config.feed_tray_location + 1].liq_in
        )

        self.feed_vap_in = Arc(
            source=self.stripping_section[
                self.config.feed_tray_location + 1].vap_out,
            destination=self.feed_tray.vap_in
        )

        self.feed_vap_out = Arc(
            source=self.feed_tray.vap_out,
            destination=self.rectification_section[
                self.config.feed_tray_location - 1].vap_in
        )

    def _make_condenser_arcs(self):

        self.condenser_vap_in = Arc(
            source=self.rectification_section[1].vap_out,
            destination=self.condenser.inlet
        )

        self.condenser_reflux_out = Arc(
            source=self.condenser.reflux,
            destination=self.rectification_section[1].liq_in
        )

    def _make_reboiler_arcs(self):

        self.reboiler_liq_in = Arc(
            source=self.stripping_section[
                self.config.number_of_trays].liq_out,
            destination=self.reboiler.inlet
        )

        self.reboiler_vap_out = Arc(
            source=self.reboiler.vapor_reboil,
            destination=self.stripping_section[
                self.config.number_of_trays].vap_in
        )

    def propagate_stream_state(self, source=None,
                               destination=None):
        """
        This method is used during initialization to propage state values
        between any two ports.

        Args:
            source : source port
            destination : destination port
        """
        for v in source.vars:
            for i in destination.vars[v]:
                if not destination.vars[v][i].fixed:
                    destination.vars[v][i].value = \
                        value(source.vars[v][i])

    def initialize(self, state_args_feed=None, state_args_liq=None,
                   state_args_vap=None, solver=None, optarg=None,
                   outlvl=idaeslog.NOTSET):

        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        init_log.info("Begin initialization.")

        solverobj = get_solver(solver, optarg)

        feed_flags = self.feed_tray.initialize(
            solver=solver, optarg=optarg, outlvl=outlvl)

        self.propagate_stream_state(
            source=self.feed_tray.vap_out,
            destination=self.condenser.inlet)

        self.condenser.initialize(
            solver=solver, optarg=optarg, outlvl=outlvl)

        self.propagate_stream_state(
            source=self.feed_tray.liq_out,
            destination=self.reboiler.inlet)

        self.reboiler.initialize(
            solver=solver, optarg=optarg, outlvl=outlvl)

        # initialize the rectification section
        for i in self._rectification_index:
            self.propagate_stream_state(
                source=self.condenser.reflux,
                destination=self.rectification_section[i].liq_in)
            self.propagate_stream_state(
                source=self.feed_tray.vap_out,
                destination=self.rectification_section[i].vap_in)
            if i == 1:
                rect_liq_flags = self.rectification_section[i]. \
                    initialize(hold_state_liq=True,
                               hold_state_vap=False,
                               solver=solver,
                               optarg=optarg,
                               outlvl=outlvl)
            elif i == len(self._rectification_index):
                rect_vap_flags = \
                    self.rectification_section[i]. \
                    initialize(hold_state_liq=False,
                               hold_state_vap=True,
                               solver=solver,
                               optarg=optarg,
                               outlvl=outlvl)
            else:
                self.rectification_section[i].initialize(
                    solver=solver, optarg=optarg, outlvl=outlvl)

        # initialize the stripping section
        for i in self._stripping_index:
            self.propagate_stream_state(
                source=self.feed_tray.liq_out,
                destination=self.stripping_section[i].liq_in)
            self.propagate_stream_state(
                source=self.reboiler.vapor_reboil,
                destination=self.stripping_section[i].vap_in)
            if i == self.config.feed_tray_location + 1:
                strip_liq_flags = self.stripping_section[i]. \
                    initialize(hold_state_liq=True,
                               hold_state_vap=False,
                               solver=solver,
                               optarg=optarg,
                               outlvl=outlvl)
            elif i == self.config.number_of_trays:
                strip_vap_flags = self.stripping_section[i]. \
                    initialize(hold_state_liq=False,
                               hold_state_vap=True,
                               solver=solver,
                               optarg=optarg,
                               outlvl=outlvl)
            else:
                self.stripping_section[i].initialize(
                    solver=None, optarg=optarg, outlvl=outlvl)

        # For initialization purposes and to enable solving individual sections
        # creating a temp block. Note that this temp block is a reference to
        # the rectification, stripping, and feed sections. Also, expanded arcs
        # are added to the temp block as the initialization solve proceeds.
        self._temp_block = Block()

        self._temp_block.rectification = Block()

        # adding reference to the rectification section and the expanded
        # vapor and liquid arcs
        self._temp_block.rectification.trays = Reference(
            self.rectification_section)
        self._temp_block.rectification.expanded_liq_stream = Reference(
            self.rectification_liq_stream[:].expanded_block)
        self._temp_block.rectification.expanded_vap_stream = Reference(
            self.rectification_vap_stream[:].expanded_block)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self._temp_block.rectification, tee=slc.tee)
        init_log.info(
            "Rectification section initialization status {}.".
            format(idaeslog.condition(res))
        )

        self._temp_block.stripping = Block()

        # adding reference to the stripping section and the expanded
        # vapor and liquid arcs
        self._temp_block.stripping.trays = Reference(self.stripping_section)
        self._temp_block.stripping.expanded_liq_stream = Reference(
            self.stripping_liq_stream[:].expanded_block)
        self._temp_block.stripping.expanded_vap_stream = Reference(
            self.stripping_vap_stream[:].expanded_block)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self._temp_block.stripping, tee=slc.tee)
        init_log.info(
            "Stripping section initialization status {}."
            .format(idaeslog.condition(res))
        )

        # releasing the fixed inlets for the vap in to the rectification
        # to enable connection with the feed tray vap out
        self.rectification_section[len(self._rectification_index)]. \
            properties_in_vap. \
            release_state(flags=rect_vap_flags, outlvl=outlvl)

        # releasing the fixed inlets for the liq in to the stripping
        # to enable connection with the feed tray liq out
        self.stripping_section[self.config.feed_tray_location + 1]. \
            properties_in_liq. \
            release_state(flags=strip_liq_flags, outlvl=outlvl)

        # Adding the feed tray to temp block solve
        self._temp_block.feed_tray = Reference(self.feed_tray)
        self._temp_block.expanded_feed_liq_stream_in = Reference(
            self.feed_liq_in.expanded_block)
        self._temp_block.expanded_feed_liq_stream_out = Reference(
            self.feed_liq_out.expanded_block)
        self._temp_block.expanded_feed_vap_stream_in = Reference(
            self.feed_vap_in.expanded_block)
        self._temp_block.expanded_feed_vap_stream_out = Reference(
            self.feed_vap_out.expanded_block)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self._temp_block, tee=slc.tee)
        init_log.info(
            "Column section initialization status {}."
            .format(idaeslog.condition(res))
        )

        self.rectification_section[1]. \
            properties_in_liq. \
            release_state(flags=rect_liq_flags, outlvl=outlvl)

        # Adding the condenser to the temp block solve
        self._temp_block.condenser = Reference(self.condenser)
        self._temp_block.expanded_condenser_vap_in = Reference(
            self.condenser_vap_in.expanded_block)
        self._temp_block.expanded_condenser_reflux_out = Reference(
            self.condenser_reflux_out.expanded_block)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self._temp_block, tee=slc.tee)
        init_log.info(
            "Column section + Condenser initialization status {}."
            .format(idaeslog.condition(res))
        )

        self.stripping_section[self.config.number_of_trays]. \
            properties_in_vap. \
            release_state(flags=strip_vap_flags, outlvl=outlvl)

        # Delete the _temp_block as next solve is solving the entire column.
        # If we add the reboiler to the temp block, it will be similar to
        # solving the original block. This ensures that if
        # initialize is triggered twice, there is no implicit replacing
        # component error.
        self.del_component(self._temp_block)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self, tee=slc.tee)
        init_log.info(
            "Column section + Condenser + Reboiler initialization status {}."
            .format(idaeslog.condition(res))
        )

        # release feed tray state once initialization is complete
        self.feed_tray.properties_in_feed.\
            release_state(flags=feed_flags, outlvl=outlvl)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Common Surrogate interface for IDAES.
"""
from pathlib import Path
from typing import Dict
import yaml
from pyomo.common.config import ConfigBlock, ConfigValue, ConfigList
import os.path, pickle


# from mypy_extensions import TypedDict

class Metrics:
    """

    Names for known types of metrics.

    Use these as keys in dictionaries, e.g.:
    m = {Metrics.RMSE: self.rmse}
    When adding attributes to this class, please include a comment with the
    prefix "#:" immediately above it, so Sphinx knows that this is documentation
    for the attribute.

    """

    #: Root mean-squared error
    RMSE = "RMSE"

    #: Mean-squared error
    MSE = "MSE"

    #: Sum of squared error
    SSE = "SSE"

    #: Time
    Time = "Time"

    #: Order
    Order = "Order"

    #: R-squared
    R2 = "R2"


class ConfigurationError(Exception):
    """
    IDAES exception to be used when configuration arguments are incorrect
    or inconsistent.
    """
    pass


# Single Surrogate Modeler
class Surrogate:
    CONFIG = ConfigBlock()

    # Common Declarations
    # CONFIG.declare('overwrite', ConfigValue(default=None, domain=bool))
    # CONFIG.declare('fname', ConfigValue(default=None, domain=str))

    def __init__(self, **settings):
        """
        Initialization fo the Surrogate Class.
        Keyword Args:
            settings            : Dictionary of user-defined configurations and settings for the selected surrogate model tool(s).
        Returns:
            *self** object containing all the input information and run results, including:
                - **self.config** (Dict)                                    : The configuration of the selected tool
                - **self._results** (Tuple)                                 : Performance metrics of the surrogate(s) trained
                - **self._model**   (Pyomo Expression)                      : Pyomo representation of resulting surrogate
                - **self._r_data_in**, **self._r_data_out** (NumPy Array)   : Sample points and output values used in training the surrogate
                - **self._v_data_in**, **self._v_data_out** (NumPy Array)   : Validation/test sample points and their true output values
                - **self.pkl_info** (Python Object)                         : Python object containing relevant surrogate model information.
        """

        # Config
        self.config = self.CONFIG(settings)
        self.modeler = None

        # Results
        self._results = {}
        self._metrics = None
        self._model = None
        self._b_built = False  # flag for regression

        # Data
        self._rdata_in = None
        self._rdata_out = None
        self._vdata_in = None
        self._vdata_out = None

        self.pkl_info = None

    def modify_config(self, **kwargs):
        """
        The ``modify_config`` method allows users to define a new surrogate instance simply by modifying one or more of the
        settings or keywords of a previously-defined surrogate instance.
        The values of the new keywords defined in **settings will directly replace the previous values stored in self.CONFIG.
        All other keywords remain unchanged.
        Args:
            **kwargs (dict)             : Dictionary containing (key, val) entries for the settings to be modified in self.CONFIG
        """
        _b_built = False
        self.config = self.CONFIG(kwargs)

    # Build

    def build_model(self):
        """
        The ``build_model`` method trains a surrogate model to an input dataset.
        It calls the core method which is called during surrogate generation: ``build_model`` sets up the surrogate problem,
        trains the surrogate, computes the metrics, creates a results object and generates the Pyomo representation of the model.
        It accepts no user input, inheriting the information passed in class initialization.
        """
        self._b_built = True

        self.pkl_info = {'In data': self._rdata_in.tolist(),
                         'Out data': self._rdata_out.tolist()}

        pass

    def update_model(self):
        """
        The ``update_model`` trains a new surrogate model based on the updated configuration/set-up defined by
        calling ``modify_config``
        It accepts no user input, inheriting the information provided in ``modify_config``.
        """
        self.build_model()
        pass

    # def generate_expression(self, variable_list): # TODO
    #     pass

    # Get Results

    def get_model(self):  # Pyomo Expression
        """
        The ``get_model`` method returns the result of the surrogate training process as a Pyomo Expression
        Returns:
            Pyomo Expression    : Pyomo expression of surrogate model trained.
        """
        return self._model


    def get_results(self):  # Metrics Object
        """
        The ``get_results`` method returns the performance metrics of the surrogate model(s) generated.
        Returns:
            Metrics of the surrogates based on the training data, including but not limited to the:
                - root mean squared error (RMSE),
                - mean squared error (MSE),
                - :math:`R^{2}`of the surrogate fit, and
                - model run time.
        """
        return self._results

    # Data Handling

    def get_regressed_data(self):
        """
        The ``get_regressed_data`` method returns the input data used in regressing the surrogate model.
        Returns:
            Tuple        : Tuple of two elements containing r_in (samples) and r_out (output values).

        """
        return (self._rdata_in, self._rdata_out)

    def regressed_data(self, r_in, r_out):  # 2D Numparray
        """
        The ``regressed_data`` method initializes the Surrogate class with the data for training the surrogate model.
        Args:
            r_in  (NumPy Array)  : Two-dimensional NumPy Array containing the samples/features.
            r_out (NumPy Array)  : Two-dimensional NumPy Array containing the output values.
        """
        self._rdata_in = r_in
        self._rdata_out = r_out

    def get_validation_data(self):
        """
        The ``get_validation_data`` method returns the data supplied for validating the surrogate model.
        Returns:
            Tuple        : Tuple of two elements containing validation data v_in (samples) and v_out (output values).

        """
        return (self._vdata_in, self._vdata_out)

    def validation_data(self, v_in, v_out):  # 2D Numparray
        """
        The ``validation_data`` method initializes the Surrogate class with data for validating/testing the surrogate model after generation.
        Args:
            v_in  (NumPy Array)  : Two-dimensional NumPy Array containing the validation samples/features.
            v_out (NumPy Array)  : Two-dimensional NumPy Array containing the output values of the validation samples.
        """
        self._vdata_in = v_in
        self._vdata_out = v_out

    # Using regressed model
    # PYLINT-TODO: check if adding self as arg to fix pylint "undefined-variable 'self'" is valid
    def calculate_outputs(self, inputs):  # 2D Numparray, use pyomo expression
        """
        ``calculate_outputs`` evaluates the output predictions from the surrogate for an array of input samples **inputs**
        Args:
            inputs(NumPy Array)     : Two-dimensional NumPy Array containing the sample points to be evaluated.
        Returns:
            outputs(NumPy Array)    : NumPy Array containing the output predictions from the surrogate model.
        """
        outputs = self._model(inputs)
        return outputs

    # Additional Metrics - MUST NOT OVERWRITE MODELER METRICS

    # def get_trained_metrics():  # 2D Nparray, 2D Numparray # TODO
    #     if not _b_built:
    #         print("Warning: No surrogate model regressed.")
    #         return
    #     pass

    # PYLINT-TODO: check if adding as an argument and using self._b_built in the body is valid
    def get_validated_metrics(self, xval, zval):  # 2D, 2D Numparray, use pyomo expression
        """
        ``get_validated_metrics`` evaluates the performance metrics for the surrogate model based on a set of off-design points (xval, zval)
        Args:
            xval (NumPy Array)      : Two-dimensional array containing a set of off-design samples.
            zval (NumPy Array)      : Array containing a true output values at off-design sample points.
        Returns:
            Metrics of the surrogate model evaluated based on the off-design data set (xval, zval), including but not limited to the:
                - root mean squared error (RMSE),
                - mean squared error (MSE), and
                - :math:`R^{2}`of the surrogate fit based on the off-design data points.
        """
        if not self._b_built:
            print("Warning: No surrogate model regressed.")
            return
        pass

    def save_results(self, filename, overwrite=False):
        """
        The ``save_results`` method saves the results of the surrogate run in a pickle object
        Args:
            filename (str)      : The name of the file to be saved. Must be of extension .pickle
            overwrite (bool)    : Boolean controlling whether any existing file with the same name is overwritten or not. Default is False.
        Raises:
            Exception:
                * **filename** is not a string or has the wrong extension type
             Exception:
                * A file with the name **filename** already exists and **overwrite** is False
             Exception:
                * A problem is encountered while trying to save the file.
        """
        # Ensure overwrite option, when entered, is boolean
        if not isinstance(overwrite, bool):
            raise Exception('overwrite must be boolean.')
        # Check if filename is a string with pickle extension
        if not isinstance(filename, str) or os.path.splitext(filename)[-1].lower() != '.pickle':
            raise Exception('filename must be a string with extension ".pickle". Please correct.')
        # If overwite is false, throw up error if the filename already exists in the destination folder
        if os.path.exists(filename) and overwrite is False:
            raise Exception(filename, 'already exists!.\n')
        # Try to save
        try:
            filehandler = open(filename, 'wb')
            pickle.dump(self.pkl_info, filehandler)
            print('\nResults saved in ', str(filename))
        except:
            raise Exception('File could not be saved.')

    def load_results(self, filename):
        """
        ``load_results`` loads the results of a saved run 'filename.obj'.
        Args:
            filename(str)       : Pickle object file containing previous solution to be loaded.
        Returns:
            **self** object containing set-up information and results of the run that created **filename**.
        Raises:
            Exception:
                * **filename** does not exist in the directory.
            Exception:
                * A problem is encountered while trying to load **filename**.
        """
        if os.path.exists(filename) is False:
            raise Exception(filename, 'does not exist in directory.')
        try:
            filehandler = open(filename, 'rb')
            loaded_res = pickle.load(filehandler)
            self.config = loaded_res['Run settings']
            self._results = loaded_res['Results']
            self._model = loaded_res['Expression']
            self._rdata_in = loaded_res['In data']
            self._rdata_out = loaded_res['Out data']
            print(str(filename), 'successfully loaded.')
            return
        except:
            raise Exception('File could not be loaded.')



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import logging
import os
from pathlib import Path
from shutil import rmtree


def create_module_scratch(module_name):
    """Create a scratch that is easily found (by developers) and removed, and doesn't
    clutter up the working directories.
    """
    path = Path("~").expanduser()
    for path_part in ".idaes", "_scratch", module_name:
        path = path / path_part
        if not path.exists():
            path.mkdir()
    return path


def rmtree_scratch(scratch_path):
    """Do our level best to remove all the temporary files.
    """
    try:
        rmtree(scratch_path, ignore_errors=True)
    except Exception as err:
        pass



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""Generic Helmholtz EOS Base Class
"""
__author__ = "John Eslick"

# Import Python libraries
import os
import enum

# Import Pyomo libraries
from pyomo.environ import (
    Constraint,
    Expression,
    Param,
    PositiveReals,
    Set,
    value,
    Var,
    ConcreteModel,
    units as pyunits
)
from pyomo.environ import ExternalFunction as EF
from pyomo.common.collections import ComponentSet
from pyomo.common.config import ConfigValue, In

# Import IDAES
from idaes.core import (
    StateBlock,
    StateBlockData,
    PhysicalParameterBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MaterialFlowBasis,
    LiquidPhase,
    VaporPhase,
    Phase,
    Component
)
from idaes.core.util.math import smooth_max
from idaes.core.util.exceptions import ConfigurationError
import idaes
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

# Logger
_log = idaeslog.getLogger(__name__)


def _available(shared_lib):
    """Make sure the compiled library functions are available.
    """
    return os.path.isfile(shared_lib)


# Python external function object name to ASL user function name and units of
# measure, in the function description comments delta = density/critical density
# and tau = critical temperature/temperature
_external_function_map = {
    "func_p": { # pressure as a function of delta and tau
        "fname": "p",
        "units": pyunits.kPa,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_u": { # internal energy as a function fo delta and tau
        "fname": "u",
        "units": pyunits.kJ/pyunits.kg,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_s": { # entropy as a function of delta and tau
        "fname": "s",
        "units": pyunits.kJ/pyunits.kg/pyunits.K,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_h": { # enthalpy as a function of delta and tau
        "fname": "h",
        "units": pyunits.kJ/pyunits.kg,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_hvpt": { # vapor enthalpy as a function of pressure and tau
        "fname": "hvpt",
        "units": pyunits.kJ/pyunits.kg,
        "arg_units": [pyunits.kPa, pyunits.dimensionless]},
    "func_hlpt": { # liquid enthalpy as a function of pressure and tau
        "fname": "hlpt",
        "units": pyunits.kJ/pyunits.kg,
        "arg_units": [pyunits.kPa, pyunits.dimensionless]},
    "func_svpt": { # vapor entropy as a function of pressure and tau
        "fname": "svpt",
        "units": pyunits.kJ/pyunits.kg/pyunits.K,
        "arg_units": [pyunits.kPa, pyunits.dimensionless]},
    "func_slpt": { # liquid entropy as a function of pressure and tau
        "fname": "slpt",
        "units": pyunits.kJ/pyunits.kg/pyunits.K,
        "arg_units": [pyunits.kPa, pyunits.dimensionless]},
    "func_uvpt": { # vapor internal energy as a function of  pressure and tau
        "fname": "uvpt",
        "units": pyunits.kJ/pyunits.kg,
        "arg_units": [pyunits.kPa, pyunits.dimensionless]},
    "func_ulpt": { # liquid internal energy as a function of  pressure and tau
        "fname": "ulpt",
        "units": pyunits.kJ/pyunits.kg,
        "arg_units": [pyunits.kPa, pyunits.dimensionless]},
    "func_tau": { # tau as a function of enthalpy and pressure
        "fname": "tau",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.kJ/pyunits.kg, pyunits.kPa]},
    "memo_test_tau": { # for memoization testing only, see func_tau
        "fname": "memo_test_tau",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.kJ/pyunits.kg, pyunits.kPa]},
    "func_tau_sp": { # tau as a function of entropy and pressure
        "fname": "tau_sp",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.kJ/pyunits.kg/pyunits.K, pyunits.kPa]},
    "func_tau_up": { # tau as a function of internal energy and pressure
        "fname": "tau_up",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.kJ/pyunits.kg, pyunits.kPa]},
    "func_p_stau": { # pressure as a function of entropy and tau
        "fname": "p_stau",
        "units": pyunits.kPa,
        "arg_units": [pyunits.kJ/pyunits.kg/pyunits.K, pyunits.dimensionless]},
    "func_vf": { # vapor fraction as a function of enthalpy and pressure
        "fname": "vf",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.kJ/pyunits.kg, pyunits.kPa]},
    "func_vfs": {  # vapor fraction as a function of entropy and pressure
        "fname": "vfs",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.kJ/pyunits.kg/pyunits.K, pyunits.kPa]},
    "func_vfu": { # vapor fraction as a function of internal energy and pressure
        "fname": "vfu",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.kJ/pyunits.kg, pyunits.kPa]},
    "func_g": { # Gibbs free energy as a function of delta and tau
        "fname": "g",
        "units": pyunits.kJ/pyunits.kg,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_f": { # Helmholtz free energy as a function of delta and tau
        "fname": "f",
        "units": pyunits.kJ/pyunits.kg,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_cv": { # Constant volume heat capacity as a function of delta and tau
        "fname": "cv",
        "units": pyunits.kJ/pyunits.kg/pyunits.K,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_cp": { # Constant pressure heat capacity as a function of delta and tau
        "fname": "cp",
        "units": pyunits.kJ/pyunits.kg/pyunits.K,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_w": { # Speed of sound as a function of delta and tau
        "fname": "w",
        "units": pyunits.m/pyunits.s,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_delta_liq": { # Liquid delta as a function of pressure and tau
        "fname": "delta_liq",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.kPa, pyunits.dimensionless]},
    "func_delta_vap": { # vapor delta as a function of pressure and tau
        "fname": "delta_vap",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.kPa, pyunits.dimensionless]},
    "func_delta_sat_l": { # saturated liquid delta as a function of tau
        "fname": "delta_sat_l",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless]},
    "func_delta_sat_v": { # saturated vapor delta as a function of tau
        "fname": "delta_sat_v",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless]},
    "func_p_sat": { # saturation pressure as a function of tau
        "fname": "p_sat",
        "units": pyunits.kPa,
        "arg_units": [pyunits.dimensionless]},
    "func_tau_sat": { # saturation tau as a function of pressure
        "fname": "tau_sat",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.kPa]},
    # phi functions, phi is dimensionless Helmholtz free energy ((fM)/(RT)).
    # underscores after phi represent partial derivatives
    "func_phi0": { # ideal part of phi as a function of delta and tau
        "fname": "phi0",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_phi0_delta": { # phi0_delta(delta)
        "fname": "phi0_delta",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless]},
    "func_phi0_delta2": { # phi0_delta_delta(delta)
        "fname": "phi0_delta2",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless]},
    "func_phi0_tau": { # phi0_tau(tau)
        "fname": "phi0_tau",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless]},
    "func_phi0_tau2": { # phi0_tau_tau(tau)
        "fname": "phi0_tau2",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless]},
    "func_phir": { # residual part of phi as a fuction of delta and tau
        "fname": "phir",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_phir_delta": { # phir_delta(delta, tau)
        "fname": "phir_delta",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_phir_delta2": { # phir_delta_delta(delta, tau)
        "fname": "phir_delta2",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_phir_tau": { # phir_tau(delta, tau)
        "fname": "phir_tau",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_phir_tau2": { # phir_tau_tau(delta, tau)
        "fname": "phir_tau2",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
    "func_phir_delta_tau": { # phir_delta_tau(delta, tau)
        "fname": "phir_delta_tau",
        "units": pyunits.dimensionless,
        "arg_units": [pyunits.dimensionless, pyunits.dimensionless]},
}


def _add_external_functions(blk, eos_tag, shared_lib, names=None):
    """Create ExternalFunction components."""
    def _fnc(x):
        x = "_".join([x, eos_tag])
        return x
    if names is None:
        names = _external_function_map.keys()
    for name in names:
        if hasattr(blk, name):
            continue
        setattr(
            blk,
            name,
            EF(library=shared_lib,
               function=_fnc(_external_function_map[name]["fname"]),
               units=_external_function_map[name]["units"],
               arg_units=_external_function_map[name]["arg_units"])
        )


class StateVars(enum.Enum):
    """
    State variable set options
    """

    PH = 1  # Pressure-Enthalpy
    TPX = 2  # Temperature-Pressure-Quality


class PhaseType(enum.Enum):
    """
    Ways to present phases to the framework
    """

    MIX = 1  # Looks like a single phase called mixed with a vapor fraction
    LG = 2  # Looks like two phases vapor and liquid
    L = 3  # Assume only liquid is present
    G = 4  # Assume only vapor is pressent


def _htpx(
    T=None, prop=None, P=None, x=None,
    Tmin=200*pyunits.K, Tmax=1200*pyunits.K,
    Pmin=1e-3*pyunits.kPa, Pmax=1e6*pyunits.kPa):
    """
    Convenience function to calculate enthalpy from temperature and either
    pressure or vapor fraction. This function can be used for inlet streams and
    initialization where temperature is known instead of enthalpy.
    User must provide values for one of these sets of values: {T, P}, {T, x},
    or {P, x}.
    Args:
        T: Temperature (between Tmin and Tmax)
        Tmin: Lower bound on allowed temperatures
        Tmax: Upper bound on allowed temperatures
        Pmin: Lower bound on allowed pressures
        Pmax: Upper bound on allowed pressures
        prop: Property block to use for the enthalpy calcuations
        P: Pressure (between Pmin and Pmax), None if saturated
        x: Vapor fraction [mol vapor/mol total] (between 0 and 1), None if
        superheated or subcooled
    Returns:
        Total molar enthalpy [J/mol].
    """
    model = ConcreteModel()
    model.param = prop.config.parameters
    te = HelmholtzThermoExpressions(model, parameters=model.param)
    te.add_funcs(["func_p_sat"])
    Tmin = pyunits.convert(Tmin, pyunits.K)
    Tmax = pyunits.convert(Tmax, pyunits.K)
    Pmin = pyunits.convert(Pmin, pyunits.Pa)
    Pmax = pyunits.convert(Pmax, pyunits.Pa)

    if not sum((P is None, T is None, x is None)) == 1:
        raise ConfigurationError(
            "htpx function must be provided exaclty two of the arguments T, P, x")
    if T is not None:
        T = pyunits.convert(T, to_units=pyunits.K)
        if not value(Tmin) <= value(T) <= value(Tmax):
            raise ConfigurationError("T = {}, ({} <= T <= {}) [K]".format(
                value(T), value(Tmin), value(Tmax)))
    if P is not None:
        P = pyunits.convert(P, to_units=pyunits.Pa)
        if not (value(Pmin) <= value(P) <= value(Pmax)):
            raise ConfigurationError("P = {}, ({} <= P <= {}) [kPa]".format(
                value(P), value(Pmin), value(Pmax)))
        if T is not None:
            # P, T may be underspecified, but assume you know it's clearly a
            # vapor of liquid so figure out which and set x.
            psat = value(pyunits.convert(
                model.func_p_sat(model.param.temperature_crit/T),
                to_units=pyunits.Pa))
            if value(P) < psat:
                x = 1
            else:
                x = 0
    if x is not None and not 0 <= x <= 1:
        raise ConfigurationError("x = {}, (0 <= x <= 1)".format(x))

    return value(te.h(T=T, p=P, x=x))


class HelmholtzThermoExpressions(object):
    """Class to write thermodynamic property expressions.  Take one of these
    possible sets of state variables: {h, p}, {u, p}, {s, p}, {s, T}, {T, x},
    {P, x}, or {T, P, x}, and return an expression for thermo property.
    This works by converting the given state varaibles and writing expressions
    for liquid and vapor density, vapor fraction, and temerature.  Then those
    can be used to calculate any other property.  You can specify an integer
    value 1 or 0 for x with any state varaibles to get liquid or vapor
    properies.
    """

    def __init__(self, blk, parameters):
        self.param = parameters
        self.blk = blk

    @staticmethod
    def _sv_str(**kwargs):
        a = [x for x in kwargs if kwargs[x] is not None]
        return ", ".join(a)

    def add_funcs(self, names=None):
        _add_external_functions(
            self.blk,
            eos_tag=self.param.eos_tag,
            shared_lib=self.param.plib,
            names=names
        )

    def basic_calculations(
            self, h=None, s=None, p=None, T=None, u=None, x=None):
        """This function is called as this basis for most thermo expression
        writer functions.  It takes the given state variables and returns
        expressions for liqid density, vapor density, vapor fraction and
        temperature, which can be used to write an expression for any thermo
        quantity.
        """
        mw = self.param.mw
        # 1.) convert units to those expected by external functions
        if h is not None:
            h *= self.param.uc_J_per_mol_to_kJ_per_kg
        if u is not None:
            u *= self.param.uc_J_per_mol_to_kJ_per_kg
        if s is not None:
            s *= self.param.uc_J_per_molK_to_kJ_per_kgK
        if p is not None:
            p *= self.param.uc_Pa_to_kPa
        if T is not None:
            tau = self.param.temperature_crit/T
        # 2.) find the block with the external functions
        blk = self.blk

        # 3.) Take given state varaibles and convert to density, T, and x

        if h is not None and p is not None:
            # h, p
            self.add_funcs(names=["func_tau", "func_vf"])
            tau = blk.func_tau(h, p)
            if x is None:
                x = blk.func_vf(h, p)
        elif s is not None and p is not None:
            # s, p
            self.add_funcs(names=["func_tau_sp", "func_vfs"])
            tau = blk.func_tau_sp(s, p)
            if x is None:
                x = blk.func_vfs(s, p)
        elif u is not None and p is not None:
            # u, p
            self.add_funcs(names=["func_tau_up", "func_vfu"])
            tau = blk.func_tau_up(u, p)
            if x is None:
                x = blk.func_vfu(u, p)
        elif s is not None and T is not None:
            # s, p
            self.add_funcs(names=["func_p_stau", "func_vfs"])
            p = blk.func_p_stau(s, tau)
            if x is None:
                x = blk.func_vfs(s, p)
        elif x is not None and T is not None and p is not None:
            # T, P, x (okay, but I hope you know what you're doing)
            pass
        elif x is not None and p is not None:
            # x, p
            self.add_funcs(names=["func_tau_sat"])
            tau = blk.func_tau_sat(p)
        elif x is not None and T is not None:
            # x, T
            self.add_funcs(names=["func_p_sat"])
            p = blk.func_p_sat(tau)
        else:
            m = ("This choice of state variables ({}) is not yet supported."
                 .format(self._sv_str(h=h, s=s, p=p, T=T, u=u, x=x)))
            _log.error(m)
            raise NotImplementedError(m)

        # 4.) Calculate density
        self.add_funcs(names=["func_delta_liq", "func_delta_vap"])
        delta_liq = blk.func_delta_liq(p, tau)
        delta_vap = blk.func_delta_vap(p, tau)

        # 5.) From here its straight forward to calculate any property
        return blk, delta_liq, delta_vap, tau, x

    def s(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        self.add_funcs(names=["func_s"])
        s = blk.func_s(delta_liq, tau)*(1-x) + blk.func_s(delta_vap, tau)*x
        return s * self.param.uc_kJ_per_kgK_to_J_per_molK

    def h(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        self.add_funcs(names=["func_h"])
        h = blk.func_h(delta_liq, tau)*(1-x) + blk.func_h(delta_vap, tau)*x
        return h * self.param.uc_kJ_per_kg_to_J_per_mol

    def u(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        self.add_funcs(names=["func_u"])
        u = blk.func_u(delta_liq, tau)*(1-x) + blk.func_u(delta_vap, tau)*x
        return u * self.param.uc_kJ_per_kg_to_J_per_mol

    def g(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        self.add_funcs(names=["func_g"])
        g = blk.func_g(delta_liq, tau)*(1-x) + blk.func_g(delta_vap, tau)*x
        return g * self.param.uc_kJ_per_kg_to_J_per_mol

    def f(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        self.add_funcs(names=["func_f"])
        f = blk.func_f(delta_liq, tau)*(1-x) + blk.func_f(delta_vap, tau)*x
        return f * self.param.uc_kJ_per_kg_to_J_per_mol

    def p(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        self.add_funcs(names=["func_p"])
        # The following line looks a bit weird, but it is okay.  When in the
        # two-phase region the pressure for both phases is the same
        p = blk.func_p(delta_liq, tau)*(1-x) + blk.func_p(delta_vap, tau)*x
        return p * self.param.uc_kPa_to_Pa

    def v(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        v = ((1-x)/delta_liq + x/delta_vap)/self.param.dens_mass_crit*self.param.mw
        return v

    def x(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        return x

    def T(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        return self.param.temperature_crit/tau

    def tau(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        return tau

    def delta_liq(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        return delta_liq

    def rho_liq(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        return delta_liq*self.param.dens_mass_crit

    def rho_mol_liq(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        return delta_liq*self.param.dens_mass_crit/self.param.mw

    def delta_vap(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        return delta_vap

    def rho_vap(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        return delta_vap*self.param.dens_mass_crit

    def rho_mol_vap(self, **kwargs):
        blk, delta_liq, delta_vap, tau, x = self.basic_calculations(**kwargs)
        return delta_vap*self.param.dens_mass_crit/self.param.mw


class HelmholtzParameterBlockData(PhysicalParameterBlock):
    CONFIG = PhysicalParameterBlock.CONFIG()

    CONFIG.declare(
        "phase_presentation",
        ConfigValue(
            default=PhaseType.MIX,
            domain=In(PhaseType),
            description="Set the way phases are presented to models",
            doc="""Set the way phases are presented to models. The MIX option
appears to the framework to be a mixed phase containing liquid and/or vapor.
The mixed option can simplify calculations at the unit model level since it can
be treated as a single phase, but unit models such as flash vessels will not be
able to treat the phases independently. The LG option presents as two separate
phases to the framework. The L or G options can be used if it is known for sure
that only one phase is present.
**default** - PhaseType.MIX
**Valid values:** {
**PhaseType.MIX** - Present a mixed phase with liquid and/or vapor,
**PhaseType.LG** - Present a liquid and vapor phase,
**PhaseType.L** - Assume only liquid can be present,
**PhaseType.G** - Assume only vapor can be present}""",
        ),
    )

    CONFIG.declare(
        "state_vars",
        ConfigValue(
            default=StateVars.PH,
            domain=In(StateVars),
            description="State variable set",
            doc="""The set of state variables to use. Depending on the use, one
state variable set or another may be better computationally. Usually pressure
and enthalpy are the best choice because they are well behaved during a phase
change.
**default** - StateVars.PH
**Valid values:** {
**StateVars.PH** - Pressure-Enthalpy,
**StateVars.TPX** - Temperature-Pressure-Quality}""",
        ),
    )

    CONFIG.declare(
        "temperature_bounds",
        ConfigValue(
            default=None,
            domain=tuple,
            description="Tuple providing default temperature bounds in the form "
                        "(lb, ub) the bounds should have units attached.",
            doc="""This is an optional tuple providing default temperature bounds.
The elements of the tuple should include units of temperature, for example, if
pyomo units is imported as pyunits, (270*pyunits.K, 1000*pyunits.K).
"""
        ),
    )

    CONFIG.declare(
        "pressure_bounds",
        ConfigValue(
            default=None,
            domain=tuple,
            description="Tuple providing default pressure bounds in the form "
                        "(lb, ub) the bounds should have units attached.",
            doc="""This is an optional tuple providing default pressure bounds.
The elements of the tuple should include units of pressure, for example, if
pyomo units is imported as pyunits, (1*pyunits.kPa, 1e6*pyunits.kPa).
"""
        ),
    )

    CONFIG.declare(
        "enthalpy_mol_bounds",
        ConfigValue(
            default=None,
            domain=tuple,
            description="Tuple providing default enthalpy per mole bounds in the"
                        " form (lb, ub) the bounds should have units attached.",
            doc="""This is an optional tuple providing default enthalpy per mole
bounds. The elements of the tuple should include units of energy per mole, for
example, if pyomo units is imported as pyunits,
(1*pyunits.J/pyunits.mol, 1e5*pyunits.J/pyunits.mol).
"""
        ),
    )

    CONFIG.declare(
        "enthalpy_mass_bounds",
        ConfigValue(
            default=None,
            domain=tuple,
            description="Tuple providing default enthalpy per mass bounds in the"
                        " form (lb, ub) the bounds should have units attached.",
            doc="""This is an optional tuple providing default enthalpy per mass
bounds. The elements of the tuple should include units of energy per mass, for
example, if pyomo units is imported as pyunits,
(1*pyunits.J/pyunits.kg, 1e5*pyunits.J/pyunits.kg).
"""
        ),
    )

    def _set_parameters(
        self,
        library,
        state_block_class,
        component_list,
        phase_equilibrium_idx,
        phase_equilibrium_list,
        mw,
        temperature_crit,
        pressure_crit,
        dens_mass_crit,
        specific_gas_constant,
        pressure_bounds,
        temperature_bounds,
        enthalpy_bounds,
        eos_tag,
        pressure_value=1e5,
        temperature_value=300,
        enthalpy_value=1000,
    ):
        """This function sets the parameters that are required for a Helmholtz
        equation of state parameter block, and ensures that all required
        parameters are set.
        """
        # Location of the *.so or *.dll file for external functions
        self.plib = library
        self.eos_tag = eos_tag
        self._state_block_class = state_block_class
        self.component_list = component_list
        self.phase_equilibrium_idx = phase_equilibrium_idx
        self.phase_equilibrium_list = phase_equilibrium_list
        # Parameters, these should match what's in the C code
        self.temperature_crit = temperature_crit
        self.pressure_crit = pressure_crit
        self.dens_mass_crit = dens_mass_crit
        self.specific_gas_constant = specific_gas_constant
        self.mw = mw
        self.default_pressure_bounds = pressure_bounds
        self.default_enthalpy_bounds = enthalpy_bounds
        self.default_temperature_bounds = temperature_bounds
        self.default_pressure_value = pressure_value
        self.default_temperature_value = temperature_value
        self.default_enthalpy_value = enthalpy_value

    def build(self):
        super().build()
        # Location of the *.so or *.dll file for external functions
        # Phase list
        self.available = _available(self.plib)

        #if state var bounds are provided in config, override the defaults
        if self.config.temperature_bounds is not None:
            self.default_temperature_bounds = (
                pyunits.convert(self.config.temperature_bounds[0], pyunits.K),
                pyunits.convert(self.config.temperature_bounds[1], pyunits.K)
            )
        if self.config.pressure_bounds is not None:
            self.default_pressure_bounds = (
                pyunits.convert(self.config.pressure_bounds[0], pyunits.Pa),
                pyunits.convert(self.config.pressure_bounds[1], pyunits.Pa)
            )
        if self.config.enthalpy_mol_bounds is not None:
            u = pyunits.J/pyunits.mol
            self.default_enthalpy_bounds = (
                pyunits.convert(self.config.enthalpy_mol_bounds[0], u),
                pyunits.convert(self.config.enthalpy_mol_bounds[1], u)
            )
        if self.config.enthalpy_mass_bounds is not None:
            u = pyunits.J/pyunits.mol
            lb = self.config.enthalpy_mass_bounds[0]*self.mw
            ub = self.config.enthalpy_mass_bounds[1]*self.mw
            self.default_enthalpy_bounds = (
                value(pyunits.convert(lb, u)),
                value(pyunits.convert(ub, u))
            )

        # Create Component objects
        for c in self.component_list:
            setattr(self, str(c), Component(
                default={"_component_list_exists": True}))

        # Create Phase objects
        self.private_phase_list = Set(initialize=["Vap", "Liq"])
        if self.config.phase_presentation == PhaseType.MIX:
            self.Mix = Phase()

        if self.config.phase_presentation == PhaseType.LG or \
                self.config.phase_presentation == PhaseType.L:
            self.Liq = LiquidPhase()

        if self.config.phase_presentation == PhaseType.LG or \
                self.config.phase_presentation == PhaseType.G:
            self.Vap = VaporPhase()

        # State var set
        self.state_vars = self.config.state_vars

        self.smoothing_pressure_over = Param(
            mutable=True,
            initialize=1e-4,
            doc="Smooth max parameter (pressure over)",
            units=pyunits.kPa)

        self.smoothing_pressure_under = Param(
            mutable=True,
            initialize=1e-4,
            doc="Smooth max parameter (pressure under)",
            units=pyunits.kPa)

        # Unit conversion factors. Currently assume everything except external
        # functions are on a molar basis with base SI units.  Most of these are
        # used to convert quanties related to external functions.  Not using
        # pyunits convert since these conversions are just prefixes.  Mass
        # to moles is a simple conversions, since this is only one component
        self.uc_J_per_mol_to_kJ_per_kg = pyunits.kJ/self.mw/1000/pyunits.J
        self.uc_kJ_per_kg_to_J_per_mol = self.mw*pyunits.J*1000/pyunits.kJ
        # Entropy type units are same as energy type units since K is left as is
        self.uc_J_per_molK_to_kJ_per_kgK = self.uc_J_per_mol_to_kJ_per_kg
        self.uc_kJ_per_kgK_to_J_per_molK = self.uc_kJ_per_kg_to_J_per_mol
        # Pressure conversions
        self.uc_kPa_to_Pa = 1000*pyunits.Pa/pyunits.kPa
        self.uc_Pa_to_kPa = pyunits.kPa/pyunits.Pa/1000

        # TODO<jce> leaving flow to not break things, but plan to remove
        self.set_default_scaling("flow_mol", 1e-4)
        self.set_default_scaling("flow_mol_comp", 1e-4)
        self.set_default_scaling("flow_vol", 100)
        self.set_default_scaling("flow_mass", 1)

        # Set some scalings with reasonable a priori values
        self.set_default_scaling("temperature_crit", 1e-2)
        self.set_default_scaling("enth_mol", 1e-3)
        self.set_default_scaling("temperature", 1e-1)
        self.set_default_scaling("pressure", 1e-6)
        self.set_default_scaling("vapor_frac", 1e1)
        self.set_default_scaling("dens_mass_phase", 1e-2, index="Liq")
        self.set_default_scaling("dens_mass_phase", 1e1)  # Not liq
        self.set_default_scaling("pressure_crit", 1e-6)
        self.set_default_scaling("dens_mass_crit", 1e-2)
        self.set_default_scaling("mw", 1e3)
        self.set_default_scaling("temperature_sat", 1e-2)
        self.set_default_scaling("temperature_red", 1)
        self.set_default_scaling("pressure_sat", 1e-5)
        self.set_default_scaling("dens_mass_phase", 1e-2, index="Liq")
        self.set_default_scaling("dens_mass_phase", 1e1)
        self.set_default_scaling("dens_phase_red", 1)
        self.set_default_scaling("enth_mol_sat_phase", 1e-2, index="Liq")
        self.set_default_scaling("enth_mol_sat_phase", 1e-4, index="Vap")
        self.set_default_scaling("dh_vap_mol", 1e-4)
        self.set_default_scaling("energy_internal_mol_phase", 1e-2, index="Liq")
        self.set_default_scaling("energy_internal_mol_phase", 1e-4)
        self.set_default_scaling("enth_mol_phase", 1e-2, index="Liq")
        self.set_default_scaling("enth_mol_phase", 1e-4, index="Vap")
        self.set_default_scaling("entr_mol_phase", 1e-1, index="Liq")
        self.set_default_scaling("entr_mol_phase", 1e-1, index="Vap")
        self.set_default_scaling("cp_mol_phase", 1e-2, index="Liq")
        self.set_default_scaling("cp_mol_phase", 1e-2, index="Vap")
        self.set_default_scaling("cv_mol_phase", 1e-2, index="Liq")
        self.set_default_scaling("cv_mol_phase", 1e-2, index="Vap")
        self.set_default_scaling("dens_mol_phase", 1e-2, index="Liq")
        self.set_default_scaling("dens_mol_phase", 1e-4, index="Vap")
        self.set_default_scaling("phase_frac", 10)
        self.set_default_scaling("enth_mol", 1e-3)
        self.set_default_scaling("energy_internal_mol", 1e-3)
        self.set_default_scaling("entr_mol", 1e-1)
        self.set_default_scaling("cp_mol", 1e-2)
        self.set_default_scaling("cv_mol", 1e-2)
        self.set_default_scaling("dens_mass", 1)
        self.set_default_scaling("dens_mol", 1e-3)
        self.set_default_scaling("heat_capacity_ratio", 1e1)
        self.set_default_scaling("enth_mass", 1)

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "temperature_crit": {"method": None, "units": "K"},
                "pressure_crit": {"method": None, "units": "Pa"},
                "dens_mass_crit": {"method": None, "units": "kg/m^3"},
                "specific_gas_constant": {"method": None, "units": "J/kg.K"},
                "mw": {"method": None, "units": "kg/mol"},
                "temperature_sat": {"method": "None", "units": "K"},
                "flow_mol": {"method": None, "units": "mol/s"},
                "flow_mass": {"method": None, "units": "kg/s"},
                "flow_vol": {"method": None, "units": "m^3/s"},
                "temperature": {"method": None, "units": "K"},
                "pressure": {"method": None, "units": "Pa"},
                "vapor_frac": {"method": None, "units": None},
                "dens_mass_phase": {"method": None, "units": "kg/m^3"},
                "temperature_red": {"method": None, "units": None},
                "pressure_sat": {"method": None, "units": "kPa"},
                "energy_internal_mol_phase": {
                    "method": None, "units": "J/mol"},
                "enth_mol_phase": {"method": None, "units": "J/mol"},
                "entr_mol_phase": {"method": None, "units": "J/mol.K"},
                "cp_mol_phase": {"method": None, "units": "J/mol.K"},
                "cv_mol_phase": {"method": None, "units": "J/mol.K"},
                "speed_sound_phase": {"method": None, "units": "m/s"},
                "dens_mol_phase": {"method": None, "units": "mol/m^3"},
                "therm_cond_phase": {"method": None, "units": "W/m.K"},
                "visc_d_phase": {"method": None, "units": "Pa.s"},
                "visc_k_phase": {"method": None, "units": "m^2/s"},
                "phase_frac": {"method": None, "units": None},
                "flow_mol_comp": {"method": None, "units": "mol/s"},
                "energy_internal_mol": {"method": None, "units": "J/mol"},
                "enth_mol": {"method": None, "units": "J/mol"},
                "entr_mol": {"method": None, "units": "J/mol.K"},
                "cp_mol": {"method": None, "units": "J/mol.K"},
                "cv_mol": {"method": None, "units": "J/mol.K"},
                "heat_capacity_ratio": {"method": None, "units": None},
                "dens_mass": {"method": None, "units": "kg/m^3"},
                "dens_mol": {"method": None, "units": "mol/m^3"},
                "dh_vap_mol": {"method": None, "units": "J/mol"},
            }
        )

        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


class _StateBlock(StateBlock):
    """
    This class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    @staticmethod
    def _set_fixed(v, f):
        if f:
            v.fix()
        else:
            v.unfix()

    def initialize(self, *args, **kwargs):
        flags = {}
        hold_state = kwargs.pop("hold_state", False)
        state_args = kwargs.pop("state_args", None)

        for i, v in self.items():
            pp = self[i].config.parameters.config.phase_presentation
            if self[i].state_vars == StateVars.PH:
                # hold the P-H vars
                flags[i] = (v.flow_mol.fixed,
                            v.enth_mol.fixed,
                            v.pressure.fixed)

                if state_args is not None:
                    if not v.flow_mol.fixed:
                        try:
                            v.flow_mol.value = state_args["flow_mol"]
                        except KeyError:
                            pass
                    if not v.enth_mol.fixed:
                        try:
                            v.enth_mol.value = state_args["enth_mol"]
                        except KeyError:
                            pass
                    if not v.pressure.fixed:
                        try:
                            v.pressure.value = state_args["pressure"]
                        except KeyError:
                            pass

                if hold_state:
                    v.flow_mol.fix()
                    v.enth_mol.fix()
                    v.pressure.fix()

            elif self[i].state_vars == StateVars.TPX:
                # Hold the T-P-x vars
                if pp in (PhaseType.MIX, PhaseType.LG):
                    flags[i] = (
                        v.flow_mol.fixed,
                        v.temperature.fixed,
                        v.pressure.fixed,
                        v.vapor_frac.fixed,
                    )

                    if state_args is not None:
                        if not v.flow_mol.fixed:
                            try:
                                v.flow_mol.value = state_args["flow_mol"]
                            except KeyError:
                                pass
                        if not v.temperature.fixed:
                            try:
                                v.temperature.value = state_args["temperature"]
                            except KeyError:
                                pass
                        if not v.pressure.fixed:
                            try:
                                v.pressure.value = state_args["pressure"]
                            except KeyError:
                                pass
                        if not v.vapor_frac.fixed:
                            try:
                                v.vapor_frac.value = state_args["vapor_frac"]
                            except KeyError:
                                pass

                    if hold_state:
                        v.flow_mol.fix()
                        v.temperature.fix()
                        v.pressure.fix()
                        v.vapor_frac.fix()
                else:
                    flags[i] = (v.flow_mol.fixed,
                                v.temperature.fixed,
                                v.pressure.fixed)

                    if state_args is not None:
                        if not v.flow_mol.fixed:
                            try:
                                v.flow_mol.value = state_args["flow_mol"]
                            except KeyError:
                                pass
                        if not v.temperature.fixed:
                            try:
                                v.temperature.value = state_args["temperature"]
                            except KeyError:
                                pass
                        if not v.pressure.fixed:
                            try:
                                v.pressure.value = state_args["pressure"]
                            except KeyError:
                                pass

                    if hold_state:
                        v.flow_mol.fix()
                        v.temperature.fix()
                        v.pressure.fix()

        # Call initialize on each data element
        for i in self:
            self[i].initialize(*args, **kwargs)
        return flags

    def release_state(self, flags, **kwargs):
        for i, f in flags.items():
            pp = self[i].config.parameters.config.phase_presentation
            if self[i].state_vars == StateVars.PH:
                self._set_fixed(self[i].flow_mol, f[0])
                self._set_fixed(self[i].enth_mol, f[1])
                self._set_fixed(self[i].pressure, f[2])
            elif self[i].state_vars == StateVars.TPX:
                self._set_fixed(self[i].flow_mol, f[0])
                self._set_fixed(self[i].temperature, f[1])
                self._set_fixed(self[i].pressure, f[2])
                if pp in (PhaseType.MIX, PhaseType.LG):
                    self._set_fixed(self[i].vapor_frac, f[3])


class HelmholtzStateBlockData(StateBlockData):
    """
    This is a base clase for Helmholtz equations of state using IDAES standard
    Helmholtz EOS external functions written in C++.
    """

    def initialize(self, *args, **kwargs):
        # With this particualr property pacakage there is not need for
        # initialization
        pass

    def _external_functions(self):
        """Create ExternalFunction components.  This includes some external
        functions that are not usually used for testing purposes."""
        _add_external_functions(
            blk=self,
            eos_tag=self.config.parameters.eos_tag,
            shared_lib=self.config.parameters.plib
        )

    def _state_vars(self):
        """ Create the state variables
        """
        params = self.config.parameters

        self.flow_mol = Var(
            initialize=1,
            doc="Total flow",
            units=pyunits.mol/pyunits.s
        )

        if self.state_vars == StateVars.PH:
            self.pressure = Var(
                domain=PositiveReals,
                initialize=params.default_pressure_value,
                doc="Pressure",
                bounds=params.default_pressure_bounds,
                units=pyunits.Pa)
            self.enth_mol = Var(
                initialize=params.default_enthalpy_value,
                doc="Total molar enthalpy",
                bounds=params.default_enthalpy_bounds,
                units=pyunits.J/pyunits.mol)

            # P is in kPa and h_mass is in kJ/kg for external func calls
            P = self.pressure * params.uc_Pa_to_kPa
            h_mass = self.enth_mol * params.uc_J_per_mol_to_kJ_per_kg
            phase_set = params.config.phase_presentation

            self.temperature = Expression(
                expr=self.temperature_crit / self.func_tau(h_mass, P),
                doc="Temperature")

            if phase_set == PhaseType.MIX or phase_set == PhaseType.LG:
                self.vapor_frac = Expression(
                    expr=self.func_vf(h_mass, P),
                    doc="Vapor mole fraction (mol vapor/mol total)")
            elif phase_set == PhaseType.L:
                self.vapor_frac = Expression(
                    expr=0.0,
                    doc="Vapor mole fraction (mol vapor/mol total)")
            elif phase_set == PhaseType.G:
                self.vapor_frac = Expression(
                    expr=1.0,
                    doc="Vapor mole fraction (mol vapor/mol total)",
                )

            # For variables that show up in ports specify extensive/intensive
            self.extensive_set = ComponentSet((self.flow_mol,))
            self.intensive_set = ComponentSet((self.enth_mol, self.pressure))

        elif self.state_vars == StateVars.TPX:
            self.temperature = Var(
                domain=PositiveReals,
                initialize=params.default_temperature_value,
                doc="Temperature",
                bounds=params.default_temperature_bounds,
                units=pyunits.K)

            self.pressure = Var(
                domain=PositiveReals,
                initialize=params.default_pressure_value,
                doc="Pressure",
                bounds=params.default_pressure_bounds,
                units=pyunits.Pa)

            self.vapor_frac = Var(
                initialize=0.0,
                units=pyunits.dimensionless,
                doc="Vapor fraction"
                # No bounds here, since it is often (usually) on it's bound
                # and that's not the best for IPOPT
            )

            # enth_mol is defined later, since in this case it needs
            # enth_mol_phase to be defined first

            # For variables that show up in ports specify extensive/intensive
            self.extensive_set = ComponentSet((self.flow_mol,))
            self.intensive_set = ComponentSet(
                (self.temperature, self.pressure, self.vapor_frac)
            )

    def _tpx_phase_eq(self):
        # Saturation pressure
        params = self.config.parameters
        eps_pu = params.smoothing_pressure_under
        eps_po = params.smoothing_pressure_over
        priv_plist = params.private_phase_list
        plist = params.phase_list
        rhoc = params.dens_mass_crit

        # Convert presssures to kPa for external functions and nicer scaling in
        # the compimentarity-type constraints.
        P = self.pressure * params.uc_Pa_to_kPa
        Psat = self.pressure_sat * params.uc_Pa_to_kPa

        vf = self.vapor_frac
        tau = self.tau

        # Terms for determining if you are above, below, or at the Psat
        self.P_under_sat = Expression(
            expr=smooth_max(0, Psat - P, eps_pu),
            doc="pressure above Psat, 0 if liqid exists [kPa]",
        )
        self.P_over_sat = Expression(
            expr=smooth_max(0, P - Psat, eps_po),
            doc="pressure below Psat, 0 if vapor exists [kPa]",
        )

        # Calculate liquid and vapor density.  If the phase doesn't exist,
        # density will be calculated at the saturation or critical pressure
        def rule_dens_mass(b, p):
            if p == "Liq":
                return rhoc * self.func_delta_liq(P + self.P_under_sat, tau)
            else:
                return rhoc * self.func_delta_vap(P - self.P_over_sat, tau)

        self.dens_mass_phase = Expression(priv_plist, rule=rule_dens_mass)

        # Reduced Density (no _mass_ identifier because mass or mol is same)
        def rule_dens_red(b, p):
            return self.dens_mass_phase[p] / rhoc

        self.dens_phase_red = Expression(
            priv_plist, rule=rule_dens_red, doc="reduced density [unitless]"
        )

        # If there is only one phase fix the vapor fraction appropriately
        if len(plist) == 1:
            if "Vap" in plist:
                self.vapor_frac.fix(1.0)
            else:
                self.vapor_frac.fix(0.0)
        elif not self.config.defined_state:
            self.eq_complementarity = Constraint(
                expr=0 == (vf * self.P_over_sat - (1 - vf) * self.P_under_sat)
            )

        # eq_sat can activated to force the pressure to be the saturation
        # pressure, if you use this constraint deactivate eq_complementarity
        self.eq_sat = Constraint(expr=P / 1000.0 == Psat / 1000.0)
        self.eq_sat.deactivate()

    def build(self, *args):
        """
        Callable method for Block construction
        """
        super().build(*args)
        params = self.config.parameters

        # Check if the library is available.
        self.available = params.available
        if not self.available:
            _log.error("Library file '{}' not found. Was it installed?".format(
                    params.plib
                )
            )

        # Add external functions
        self._external_functions()

        # Thermo expression writer
        te = HelmholtzThermoExpressions(
            blk=self,
            parameters=params
        )

        # Which state vars to use
        self.state_vars = params.state_vars
        # The private phase list contains phases that may be present and is
        # used internally.  If using the single mixed phase option the phase
        # list would be mixed while the private phase list would be
        # ["Liq, "Vap"]
        phlist = params.private_phase_list
        pub_phlist = params.phase_list
        component_list = params.component_list
        phase_set = params.config.phase_presentation
        self.phase_equilibrium_list = \
            params.phase_equilibrium_list

        # Expressions that link to some parameters in the param block, which
        # are commonly needed, this lets you get the parameters with scale
        # factors directly from the state block
        self.temperature_crit = Expression(
            expr=params.temperature_crit)
        self.pressure_crit = Expression(
            expr=params.pressure_crit)
        self.dens_mass_crit = Expression(
            expr=params.dens_mass_crit)
        self.mw = Expression(
            expr=params.mw, doc="molecular weight [kg/mol]"
        )

        # create the appropriate state variables
        self._state_vars()

        # Some parameters/variables show up in several expressions, so to
        # enhance readability and compactness, give them short aliases
        Tc = params.temperature_crit
        rhoc = params.dens_mass_crit
        mw = self.mw
        # P is pressure in kPa for external function calls
        P = self.pressure * params.uc_Pa_to_kPa
        T = self.temperature
        vf = self.vapor_frac

        # Saturation temperature expression
        self.temperature_sat = Expression(
            expr=Tc / self.func_tau_sat(P), doc="Stauration temperature (K)"
        )

        # Saturation tau (tau = Tc/T)
        self.tau_sat = Expression(expr=self.func_tau_sat(P))

        # Reduced temperature
        self.temperature_red = Expression(
            expr=T / Tc, doc="reduced temperature T/Tc (unitless)"
        )

        self.tau = Expression(expr=Tc / T, doc="Tc/T (unitless)")
        tau = self.tau

        # Saturation pressure
        self.pressure_sat = Expression(
            expr=self.func_p_sat(tau) * params.uc_kPa_to_Pa,
            doc="Saturation pressure (Pa)"
        )

        if self.state_vars == StateVars.PH:
            # If TPx state vars the expressions are given in _tpx_phase_eq
            # Calculate liquid and vapor density.  If the phase doesn't exist,
            # density will be calculated at the saturation or critical pressure
            # depending on whether the temperature is above the critical
            # temperature supercritical fluid is considered to be the liquid
            # phase
            def rule_dens_mass(b, p):
                if p == "Liq":
                    return rhoc * self.func_delta_liq(P, tau)
                else:
                    return rhoc * self.func_delta_vap(P, tau)

            self.dens_mass_phase = Expression(
                phlist,
                rule=rule_dens_mass,
                doc="Mass density by phase (kg/m3)"
            )

            # Reduced Density (no _mass_ identifier as mass or mol is same)
            def rule_dens_red(b, p):
                return self.dens_mass_phase[p] / rhoc

            self.dens_phase_red = Expression(
                phlist, rule=rule_dens_red, doc="reduced density (unitless)"
            )

        elif self.state_vars == StateVars.TPX:
            self._tpx_phase_eq()
        delta = self.dens_phase_red

        # Phase property expressions all converted to SI

        # Saturated Enthalpy
        def rule_enth_mol_sat_phase(b, p):
            if p == "Liq":
                return (self.func_hlpt(P, self.tau_sat) *
                    params.uc_kJ_per_kg_to_J_per_mol)

            elif p == "Vap":
                return (self.func_hvpt(P, self.tau_sat) *
                    params.uc_kJ_per_kg_to_J_per_mol)

        self.enth_mol_sat_phase = Expression(
            phlist,
            rule=rule_enth_mol_sat_phase,
            doc="Saturated enthalpy of the phases at pressure (J/mol)",
        )

        self.dh_vap_mol = Expression(
            expr=(self.enth_mol_sat_phase["Vap"] -
                  self.enth_mol_sat_phase["Liq"]),
            doc="Enthaply of vaporization at pressure and saturation (J/mol)",
        )

        # Phase Internal Energy
        def rule_energy_internal_mol_phase(b, p):
            return self.func_u(delta[p], tau) * params.uc_kJ_per_kg_to_J_per_mol

        self.energy_internal_mol_phase = Expression(
            phlist,
            rule=rule_energy_internal_mol_phase,
            doc="Phase internal energy or saturated if phase doesn't exist",
        )

        # Phase Enthalpy
        def rule_enth_mol_phase(b, p):
            return self.func_h(delta[p], tau) * params.uc_kJ_per_kg_to_J_per_mol

        self.enth_mol_phase = Expression(
            phlist,
            rule=rule_enth_mol_phase,
            doc="Phase enthalpy or saturated if phase doesn't exist [J/mol]",
        )

        # Phase Entropy
        def rule_entr_mol_phase(b, p):
            return (self.func_s(delta[p], tau) *
                params.uc_kJ_per_kgK_to_J_per_molK)

        self.entr_mol_phase = Expression(
            phlist,
            rule=rule_entr_mol_phase,
            doc="Phase entropy or saturated if phase doesn't exist [J/mol/K]",
        )

        # Phase constant pressure heat capacity, cp
        def rule_cp_mol_phase(b, p):
            return (self.func_cp(delta[p], tau) *
                params.uc_kJ_per_kgK_to_J_per_molK)

        self.cp_mol_phase = Expression(
            phlist,
            rule=rule_cp_mol_phase,
            doc="Phase cp or saturated if phase doesn't exist [J/mol/K]",
        )

        # Phase constant pressure heat capacity, cv
        def rule_cv_mol_phase(b, p):
            return (self.func_cv(delta[p], tau) *
                params.uc_kJ_per_kgK_to_J_per_molK)

        self.cv_mol_phase = Expression(
            phlist,
            rule=rule_cv_mol_phase,
            doc="Phase cv or saturated if phase doesn't exist [J/mol/K]",
        )

        # Phase speed of sound
        def rule_speed_sound_phase(b, p):
            return self.func_w(delta[p], tau)

        self.speed_sound_phase = Expression(
            phlist,
            rule=rule_speed_sound_phase,
            doc="Phase speed of sound or saturated if phase doesn't exist",
        )

        # Phase Mole density
        def rule_dens_mol_phase(b, p):
            return self.dens_mass_phase[p] / mw

        self.dens_mol_phase = Expression(
            phlist,
            rule=rule_dens_mol_phase,
            doc="Phase mole density or saturated if phase doesn't exist")

        # Phase fraction
        def rule_phase_frac(b, p):
            if p == "Vap":
                return vf
            elif p == "Liq":
                return 1.0 - vf

        self.phase_frac = Expression(
            phlist, rule=rule_phase_frac, doc="Phase fraction [unitless]"
        )

        # Component flow (for units that need it)
        def component_flow(b, i):
            return self.flow_mol

        self.flow_mol_comp = Expression(
            component_list,
            rule=component_flow,
            doc="Total flow (both phases) of component [mol/s]",
        )

        # Total (mixed phase) properties

        # Enthalpy
        if self.state_vars == StateVars.TPX:
            self.enth_mol = Expression(
                expr=sum(self.phase_frac[p] * self.enth_mol_phase[p]
                         for p in phlist)
            )
        # Internal Energy
        self.energy_internal_mol = Expression(
            expr=sum(
                self.phase_frac[p] * self.energy_internal_mol_phase[p]
                for p in phlist
            )
        )
        # Entropy
        self.entr_mol = Expression(expr=te.s(h=self.enth_mol, p=self.pressure))
        # cp
        self.cp_mol = Expression(
            expr=sum(self.phase_frac[p] * self.cp_mol_phase[p]
                     for p in phlist)
        )
        # cv
        self.cv_mol = Expression(
            expr=sum(self.phase_frac[p] * self.cv_mol_phase[p] for p in phlist)
        )

        # mass density
        self.dens_mass = Expression(
            expr=1.0
            / sum(self.phase_frac[p] * 1.0 / self.dens_mass_phase[p]
                  for p in phlist)
        )
        # mole density
        self.dens_mol = Expression(
            expr=1.0
            / sum(self.phase_frac[p] * 1.0 / self.dens_mol_phase[p]
                  for p in phlist)
        )
        # heat capacity ratio
        self.heat_capacity_ratio = Expression(expr=self.cp_mol / self.cv_mol)
        # Flows
        self.flow_vol = Expression(
            expr=self.flow_mol / self.dens_mol,
            doc="Total liquid + vapor volumetric flow (m3/s)",
        )

        self.flow_mass = Expression(
            expr=self.mw * self.flow_mol, doc="mass flow rate [kg/s]"
        )

        self.enth_mass = Expression(expr=self.enth_mol / mw,
                                    doc="Mass enthalpy (J/kg)")

        # Set the state vars dictionary
        if self.state_vars == StateVars.PH:
            self._state_vars_dict = {
                "flow_mol": self.flow_mol,
                "enth_mol": self.enth_mol,
                "pressure": self.pressure,
            }
        elif self.state_vars == StateVars.TPX and phase_set in (
            PhaseType.MIX,
            PhaseType.LG,
        ):
            self._state_vars_dict = {
                "flow_mol": self.flow_mol,
                "temperature": self.temperature,
                "pressure": self.pressure,
                "vapor_frac": self.vapor_frac,
            }
        elif self.state_vars == StateVars.TPX and phase_set in (
            PhaseType.G,
            PhaseType.L,
        ):
            self._state_vars_dict = {
                "flow_mol": self.flow_mol,
                "temperature": self.temperature,
                "pressure": self.pressure,
            }

        def rule_mole_frac_phase_comp(b, p, j):
            if p == "Mix":
                return 1.0
            else:
                return self.phase_frac[p]
        self.mole_frac_phase_comp = Expression(
            pub_phlist,
            component_list,
            rule=rule_mole_frac_phase_comp
        )

        # Define some expressions for the balance terms returned by functions
        # This is just to allow assigning scale factors to the expressions
        # returned
        #
        # Marterial flow term exprsssions
        def rule_material_flow_terms(b, p):
            if p == "Mix":
                return self.flow_mol
            else:
                return self.flow_mol * self.phase_frac[p]

        self.material_flow_terms = Expression(pub_phlist,
                                              rule=rule_material_flow_terms)

        # Enthaply flow term expressions
        def rule_enthalpy_flow_terms(b, p):
            if p == "Mix":
                return self.enth_mol * self.flow_mol
            else:
                return (self.enth_mol_phase[p] * self.phase_frac[p] *
                        self.flow_mol)

        self.enthalpy_flow_terms = Expression(pub_phlist,
                                              rule=rule_enthalpy_flow_terms)

        # Energy density term expressions
        def rule_energy_density_terms(b, p):
            if p == "Mix":
                return self.dens_mol * self.energy_internal_mol
            else:
                return (self.dens_mol_phase[p] *
                        self.energy_internal_mol_phase[p])

        self.energy_density_terms = Expression(
            pub_phlist, rule=rule_energy_density_terms
        )

    def get_material_flow_terms(self, p, j):
        return self.material_flow_terms[p]

    def get_enthalpy_flow_terms(self, p):
        return self.enthalpy_flow_terms[p]

    def get_material_density_terms(self, p, j):
        if p == "Mix":
            return self.dens_mol
        else:
            return self.dens_mol_phase[p]

    def get_energy_density_terms(self, p):
        return self.energy_density_terms[p]

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def get_material_flow_basis(b):
        return MaterialFlowBasis.molar

    def define_state_vars(self):
        return self._state_vars_dict

    def define_display_vars(self):
        return {
            "Molar Flow (mol/s)": self.flow_mol,
            "Mass Flow (kg/s)": self.flow_mass,
            "T (K)": self.temperature,
            "P (Pa)": self.pressure,
            "Vapor Fraction": self.vapor_frac,
            "Molar Enthalpy (J/mol)": self.enth_mol_phase,
        }

    def extensive_state_vars(self):
        return self.extensive_set

    def intensive_state_vars(self):
        return self.intensive_set

    def model_check(self):
        pass

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        sf_flow = iscale.get_scaling_factor(self.flow_mol, default=1)
        sf_enth = iscale.get_scaling_factor(self.enth_mol, default=1)
        sf_inte = iscale.get_scaling_factor(self.energy_internal_mol,
                                            default=1)
        sf_dens = iscale.get_scaling_factor(self.dens_mol, default=1)
        sf_pres = iscale.get_scaling_factor(self.pressure, default=1)
        for v in self.material_flow_terms.values():
            iscale.set_scaling_factor(v, sf_flow)
        for v in self.enthalpy_flow_terms.values():
            iscale.set_scaling_factor(v, sf_enth*sf_flow)
        for k, v in self.energy_density_terms.items():
            if k == "Mix":
                iscale.set_scaling_factor(v, sf_inte*sf_dens)
            else:
                sf_inte_p = iscale.get_scaling_factor(
                    self.energy_internal_mol_phase[k], default=1)
                sf_dens_p = iscale.get_scaling_factor(
                    self.dens_mol_phase[k], default=1)
                iscale.set_scaling_factor(v, sf_inte_p*sf_dens_p)
        try:
            iscale.set_scaling_factor(self.eq_sat, sf_pres/1000.0)
        except AttributeError:
            pass  # may not have eq_sat, and that's ok
        try:
            iscale.set_scaling_factor(self.eq_complementarity, sf_pres/10)
        except AttributeError:
            pass  # may not have eq_complementarity which is fine



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This file builds the kriging SOFC reduced order model (developed by Pacific
Northwest National Laboratory) in a Pyomo block. A data file of kriging
coefficients must be provided. The ROM is specifically designed for use with
the NGFC flowsheet. The code for generating SOFC ROMS can be found on PNNL's
Github page: https://github.com/NGFC-Lib/NGFC-Lib.
"""

import os
from pyomo.environ import Block, Constraint, Param, Var, exp, value, units
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.common.fileutils import this_file_dir


# creates a dictionary from a list of indices and values
def build_dict(index, values):
    d = {}
    for i, v in zip(index, values):
        d[i] = v
    return d


# creates a dictionary with double indices
def build_matrix(index1, index2, values):
    d = {}
    count = 0
    for i in index1:
        for j in index2:
            d[(i, j)] = values[count]
            count += 1
    return d


def build_SOFC_ROM(m):
    m.SOFC = b = Block()

    # load kriging coefficients
    fname = 'kriging_coefficients.dat'
    path = os.path.join(this_file_dir(), fname)

    with open(path) as file:
        text = file.readlines()

    kriging = [float(line) for line in text]

    n_inputs = 9
    n_outputs = 48
    n_samples = 13424

    # create indecies for vars and params
    input_index = list(range(n_inputs))
    input_plus_index = list(range(n_inputs+1))
    output_index = list(range(n_outputs))
    samples_index = list(range(n_samples))

    # read through data file and create params
    start_count = 0
    end_count = n_inputs
    values = kriging[start_count:end_count]
    b.mean_input = Param(input_index,
                         initialize=build_dict(input_index, values),
                         mutable=False)

    start_count += n_inputs
    end_count += n_inputs
    values = kriging[start_count:end_count]
    b.sigma_input = Param(input_index,
                          initialize=build_dict(input_index, values),
                          mutable=False)

    start_count += n_inputs
    end_count += n_outputs
    values = kriging[start_count:end_count]
    b.mean_output = Param(output_index,
                          initialize=build_dict(output_index, values),
                          mutable=False)

    start_count += n_outputs
    end_count += n_outputs
    values = kriging[start_count:end_count]
    b.sigma_output = Param(output_index,
                           initialize=build_dict(output_index, values),
                           mutable=False)

    start_count += n_outputs
    end_count += n_inputs*n_samples
    values = kriging[start_count:end_count]
    b.ds_input = Param(samples_index, input_index,
                       initialize=build_matrix(
                           samples_index, input_index, values),
                       mutable=False)

    start_count += n_inputs*n_samples
    end_count += n_inputs
    values = kriging[start_count:end_count]
    b.theta = Param(input_index,
                    initialize=build_dict(input_index, values),
                    mutable=False)

    start_count += n_inputs
    end_count += (n_inputs+1)*n_outputs
    values = kriging[start_count:end_count]
    b.beta = Param(input_plus_index, output_index,
                   initialize=build_matrix(
                       input_plus_index, output_index, values),
                   mutable=False)

    start_count += (n_inputs+1)*n_outputs
    end_count += n_samples*n_outputs
    values = kriging[start_count:end_count]
    b.gamma = Param(samples_index, output_index,
                    initialize=build_matrix(
                        samples_index, output_index, values),
                    mutable=False)

    # create input vars for the user to interface with
    b.current_density = Var(initialize=4000,
                            units=units.A/units.m**2,
                            bounds=(2000, 6000))
    # units for T should be degC but pyomo doesn't support conversion from K
    b.fuel_temperature = Var(initialize=500,
                             units=None,
                             bounds=(15, 600))

    b.internal_reforming = Var(initialize=0.4,
                               units=None,
                               bounds=(0, 1))

    b.air_temperature = Var(initialize=700,
                            units=None,
                            bounds=(550, 800))

    b.air_recirculation = Var(initialize=0.5,
                              units=None,
                              bounds=(0, 0.8))

    b.OTC = Var(initialize=2.1,
                units=None,
                bounds=(1.5, 3))

    b.fuel_util = Var(initialize=0.85,
                      units=None,
                      bounds=(0.4, 0.95))

    b.air_util = Var(initialize=0.5,
                     units=None,
                     bounds=(0.125, 0.833))

    b.pressure = Var(initialize=1,
                     units=units.atm,
                     bounds=(1, 2.5))

    # create vars for intermediate calculations
    ROM_initialize_values = [4000, 500, 0.4, 700, 0.5, 2.1, .85, 0.5, 1]
    b.ROM_input = Var(input_index,
                      initialize=build_dict(input_index,
                                            ROM_initialize_values))

    b.norm_input = Var(input_index, initialize=0)

    b.F = Var(input_plus_index, initialize=0)
    b.F[0].fix(1)

    b.R = Var(samples_index, initialize=0)

    b.norm_output = Var(output_index, initialize=0)

    b.ROM_output = Var(output_index)

    # create kriging regression constraints

    # this dict maps the index values to the input vars
    input_map = {0: b.current_density,
                 1: b.fuel_temperature,
                 2: b.internal_reforming,
                 3: b.air_temperature,
                 4: b.air_recirculation,
                 5: b.OTC,
                 6: b.fuel_util,
                 7: b.air_util,
                 8: b.pressure}

    def input_rule(b, i):
        if units.get_units(input_map[i]) is None:
            return b.ROM_input[i] == input_map[i]
        else:
            unit_conversion = units.get_units(input_map[i])
            return b.ROM_input[i] == input_map[i]/unit_conversion

    b.input_mapping_eqs = Constraint(input_index, rule=input_rule)

    def norm_input_rule(b, i):
        return (b.norm_input[i] ==
                (b.ROM_input[i] - b.mean_input[i])/b.sigma_input[i])

    b.norm_input_eqs = Constraint(input_index, rule=norm_input_rule)

    def F_rule(b, i):
        return b.F[i+1] == b.norm_input[i]

    b.F_eqs = Constraint(input_index, rule=F_rule)

    def R_rule(b, i):
        return (b.R[i] == exp(-1*sum(b.theta[j] *
                                     (b.ds_input[i, j] - b.norm_input[j])**2
                                     for j in input_index)))

    b.R_eqs = Constraint(samples_index, rule=R_rule)

    def norm_output_rule(b, i):
        return (b.norm_output[i] ==
                sum(b.F[j]*b.beta[j, i] for j in input_plus_index)
                + sum(b.R[k]*b.gamma[k, i] for k in samples_index))

    b.norm_output_eqs = Constraint(output_index, rule=norm_output_rule)

    def ROM_output_rule(b, i):
        return (b.ROM_output[i] ==
                b.mean_output[i] + b.norm_output[i]*b.sigma_output[i])

    b.ROM_output_eqs = Constraint(output_index, rule=ROM_output_rule)

    # create output variables and constraints
    b.anode_outlet_temperature = Var(initialize=600, units=None)
    b.cathode_outlet_temperature = Var(initialize=600, units=None)
    b.stack_voltage = Var(initialize=1, units=units.V)
    b.max_cell_temperature = Var(initialize=750, units=None)
    b.deltaT_cell = Var(initialize=100, units=None)

    def anode_outlet_rule(b):
        return b.anode_outlet_temperature == b.ROM_output[11]
    b.anode_outlet_eq = Constraint(rule=anode_outlet_rule)

    def cathode_outlet_rule(b):
        return b.cathode_outlet_temperature == b.ROM_output[13]
    b.cathode_outlet_eq = Constraint(rule=cathode_outlet_rule)

    def stack_voltage_rule(b):
        return b.stack_voltage == b.ROM_output[1]*units.V
    b.stack_voltage_eq = Constraint(rule=stack_voltage_rule)

    def max_cell_temp_rule(b):
        return b.max_cell_temperature == b.ROM_output[8]
    b.max_cell_temp_eq = Constraint(rule=max_cell_temp_rule)

    def deltaT_cell_rule(b):
        return b.deltaT_cell == b.ROM_output[10]
    b.deltaT_cell_eq = Constraint(rule=deltaT_cell_rule)


# initialization procedure for the ROM
def initialize_SOFC_ROM(b):

    def cvfc_indexed(variable, constraint):
        for i in constraint.keys():
            calculate_variable_from_constraint(variable[i],
                                               constraint[i])

    # F needs to be done manually
    def calculate_F():
        for i in b.norm_input.keys():
            b.F[i+1].fix(value(b.norm_input[i]))
            b.F[i+1].unfix()

    print('Starting ROM initialization')

    cvfc_indexed(b.ROM_input, b.input_mapping_eqs)
    cvfc_indexed(b.norm_input, b.norm_input_eqs)
    calculate_F()
    cvfc_indexed(b.R, b.R_eqs)
    cvfc_indexed(b.norm_output, b.norm_output_eqs)
    cvfc_indexed(b.ROM_output, b.ROM_output_eqs)
    cvfc_indexed(b.anode_outlet_temperature, b.anode_outlet_eq)
    cvfc_indexed(b.cathode_outlet_temperature, b.cathode_outlet_eq)
    cvfc_indexed(b.stack_voltage, b.stack_voltage_eq)
    cvfc_indexed(b.max_cell_temperature, b.max_cell_temp_eq)
    cvfc_indexed(b.deltaT_cell, b.deltaT_cell_eq)

    print('ROM initialization completed')



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for smooth VLE formulation

Authors: Andrew Lee
"""

import pytest

from pyomo.environ import (ConcreteModel,
                           Constraint,
                           Expression,
                           Param,
                           value,
                           Var,
                           units as pyunits)

from idaes.generic_models.properties.core.generic.generic_property import \
    GenericParameterBlock
from idaes.generic_models.properties.core.state_definitions import FTPx
from idaes.generic_models.properties.core.phase_equil import SmoothVLE
from idaes.generic_models.properties.core.phase_equil.forms import fugacity
from idaes.core.util.exceptions import ConfigurationError


# Dummy EoS to use for fugacity calls
class DummyEoS(object):
    def common(self, pobj):
        pass

    def build_parameters(b):
        pass

    def fug_phase_comp(b, p, j, pp):
        return b.fug_phase_comp[p, j]


@pytest.fixture()
def frame():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = GenericParameterBlock(default={
        "components": {"H2O": {"parameter_data": {"temperature_crit": 647.3},
                               "phase_equilibrium_form":
                                   {("Vap", "Liq"): fugacity}}},
        "phases": {"Liq": {"equation_of_state": DummyEoS},
                   "Vap": {"equation_of_state": DummyEoS}},
        "state_definition": FTPx,
        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K}})

    # Create a dummy state block
    m.props = m.params.state_block_class([1], default={"parameters": m.params})

    m.props[1].temperature_bubble = Var([("Liq", "Vap")], initialize=300)
    m.props[1].temperature_dew = Var([("Liq", "Vap")], initialize=300)
    m.props[1]._teq = Var([("Liq", "Vap")], initialize=300)

    m.props[1].fug_phase_comp = Var(m.params.phase_list,
                                    m.params.component_list,
                                    initialize=10)

    SmoothVLE.phase_equil(m.props[1], ("Liq", "Vap"))

    return m


@pytest.mark.unit
def test_build(frame):
    assert isinstance(frame.props[1].eps_1_Liq_Vap, Param)
    assert value(frame.props[1].eps_1_Liq_Vap) == 0.01

    assert isinstance(frame.props[1].eps_2_Liq_Vap, Param)
    assert value(frame.props[1].eps_2_Liq_Vap) == 0.0005

    assert isinstance(frame.props[1]._t1_Liq_Vap, Var)

    assert isinstance(frame.props[1]._t1_constraint_Liq_Vap, Constraint)
    assert isinstance(frame.props[1]._teq_constraint_Liq_Vap, Constraint)


@pytest.mark.unit
def test_t1(frame):
    # Test that T1 is the max(T, T_bubble)
    # Can't check directly, but see that residual of constraint is correct
    for t in [200, 300, 400, 500]:
        for tb in [200, 300, 400, 500]:
            frame.props[1].temperature.value = t
            frame.props[1].temperature_bubble[("Liq", "Vap")].value = tb
            frame.props[1]._t1_Liq_Vap.value = max(t, tb)
            assert value(frame.props[1]._t1_constraint_Liq_Vap.body) == \
                pytest.approx(0, abs=5e-3)


@pytest.mark.unit
def test_t_eq(frame):
    # Test that Teq is the min(T1, T_dew)
    # Can't check directly, but see that residual of constraint is correct
    for t1 in [200, 300, 400, 500]:
        for td in [200, 300, 400, 500]:
            frame.props[1]._t1_Liq_Vap.value = t1
            frame.props[1].temperature_dew[("Liq", "Vap")].value = td
            frame.props[1]._teq[("Liq", "Vap")].value = min(t1, td)
            assert value(frame.props[1]._teq_constraint_Liq_Vap.body) == \
                pytest.approx(0, abs=5e-3)


@pytest.mark.unit
def test_non_VLE_pair():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = GenericParameterBlock(default={
        "components": {"H2O": {"parameter_data": {"temperature_crit": 647.3},
                               "phase_equilibrium_form":
                                   {("Sol", "Liq"): fugacity}}},
        "phases": {"Sol": {"equation_of_state": DummyEoS},
                   "Liq": {"equation_of_state": DummyEoS}},
        "state_definition": FTPx,
        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K}})

    # Create a dummy state block
    m.props = m.params.state_block_class([1], default={"parameters": m.params})

    with pytest.raises(ConfigurationError,
                       match="params Generic Property Package phase pair "
                       "Liq-Sol was set to use Smooth VLE formulation, "
                       "however this is not a vapor-liquid pair."):
        SmoothVLE.phase_equil(m.props[1], ("Liq", "Sol"))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

__author__ = "John Eslick"

import pytest
from pyomo.environ import ConcreteModel, value
import idaes.generic_models.properties.swco2 as swco2
import idaes


class TestSWCO2(object):

    @pytest.fixture(scope="class")
    def model2(self):
        model = ConcreteModel()
        model.prop_param = swco2.SWCO2ParameterBlock()
        model.prop_in = swco2.SWCO2StateBlock(
            default={"parameters": model.prop_param}
        )
        return model

    @pytest.mark.unit
    def test_transport(self, model2):
        """Transport property tests
        """

        data = (
           #(T,      P,        mu,          tc,         tol_mu,  tol_tc)
            (200,    0.1e6,    10.06e-6,    9.63e-3,    0.02,    0.02),
            (240,    0.1e6,    12.07e-6,    12.23e-3,   0.02,    0.02),
            (240,    10.0e6,   188.91e-6,   159.24e-3,  0.02,    0.02),
            (240,    25e6,     214.16e-6,   171.59e-3,  0.02,    0.02),
            (280,    0.1e6,    14.05e-6,    15.19e-3,   0.02,    0.02),
            (280,    2.5e6,    14.51e-6,    17.46e-3,   0.02,    0.04),
            (280,    30.0e6,   134.98e-6,   136.63e-3,  0.02,    0.02),
            (280,    50.0e6,   160.73e-6,   152.67e-3,  0.02,    0.02),
            (306,    7.5e6,    24.06e-6,    23.93e-3,   0.02,    0.40), # supercritcal near critical
            (400,    5.0e6,    20.37e-6,    27.46e-3,   0.02,    0.02),
            (400,    50.0e6,   65.67e-6,    86.48e-3,   0.02,    0.02),
            (400,    100e6,    101.41e-6,   121.41e-3,  0.02,    0.02),
            (600,    50e6,     42.20e-6,    64.11e-3,   0.02,    0.02),
            (1000,   5e6,      41.42e-6,    71.18e-3,   0.02,    0.02),
            (1000,   50e6,     46.16e-6,    80.33e-3,   0.02,    0.02),
            (1000,   100e6,    54.79e-6,    92.16e-3,   0.02,    0.02),
        )

        for d in data:
            tol_mu = d[4]
            tol_tc = d[5]
            P = d[1]
            T = d[0]
            mu_data = d[2]
            tc_data = d[3]
            model2.prop_in.temperature.set_value(T)
            model2.prop_in.pressure = P
            Tsat = value(model2.prop_in.temperature_sat)
            if P >= 7.377e6 and T >= 304.1282:
                # super critical, which we clasify as liquid
                ph = "Liq"
            if P <= 0.5179e6:
                # below triple point
                ph = "Vap"
            elif T > Tsat + 0.5:
                ph = "Vap"
            elif T < Tsat - 0.5:
                ph = "Liq"
            else:
                # if too close to sat, don't want to get into a situation where
                # I don't know if it's liquid or vapor, so using extreme
                # caution
                # If we're woried about it, we can add tests on the sat curve,
                # and test both liquid and vapor
                continue

            rho = value(model2.prop_in.dens_mass_phase[ph])
            mu = value(model2.prop_in.visc_d_phase[ph])
            tc = value(model2.prop_in.therm_cond_phase[ph])
            print("T = {}, P = {}, mu = {}, tc = {}, rho = {}, phase = {}".
                  format(T, P, mu, tc, rho, ph)
                  )
            print("Data: mu = {}, tc = {}".format(mu_data, tc_data))
            assert tc == pytest.approx(tc_data, rel=tol_tc)
            assert mu == pytest.approx(mu_data, rel=tol_mu)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
from idaes.power_generation.unit_models.helm.splitter import HelmSplitter
from idaes.power_generation.unit_models.helm.valve_steam import HelmValve, ValveFunctionType
from idaes.power_generation.unit_models.helm.mixer import HelmMixer
from idaes.generic_models.unit_models import MomentumMixingType
from idaes.power_generation.unit_models.helm.compressor import HelmIsentropicCompressor
from idaes.power_generation.unit_models.helm.turbine import HelmIsentropicTurbine
from idaes.power_generation.unit_models.helm.pump import HelmPump
from idaes.power_generation.unit_models.helm.turbine_inlet import HelmTurbineInletStage
from idaes.power_generation.unit_models.helm.turbine_stage import HelmTurbineStage
from idaes.power_generation.unit_models.helm.turbine_outlet import HelmTurbineOutletStage
from idaes.power_generation.unit_models.helm.turbine_multistage import HelmTurbineMultistage
from idaes.power_generation.unit_models.helm.condenser_ntu import HelmNtuCondenser, HelmNtuCondenserData



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Multistage steam turbine for power generation.

Liese, (2014). "Modeling of a Steam Turbine Including Partial Arc Admission
    for Use in a Process Simulation Software Environment." Journal of Engineering
    for Gas Turbines and Power. v136, November
"""
import copy

import pyomo.environ as pyo
from pyomo.network import Arc
from pyomo.common.config import ConfigBlock, ConfigValue, ConfigList, In

from idaes.core import declare_process_block_class, UnitModelBlockData, useDefault
from idaes.power_generation.unit_models.helm import (
    HelmSplitter,
    HelmMixer,
    MomentumMixingType,
    HelmTurbineInletStage,
    HelmTurbineStage,
    HelmTurbineOutletStage,
    ValveFunctionType,
)
from idaes.power_generation.unit_models.helm import HelmValve as SteamValve

from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.util.misc import copy_port_values as copy_port
import idaes.core.util.scaling as iscale

import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


def _define_turbine_multistage_config(config):
    config.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag",
            doc="Only False, in a dynamic flowsheet this is psuedo-steady-state.",
        ),
    )
    config.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag",
            doc="Only False, in a dynamic flowsheet this is psuedo-steady-state.",
        ),
    )
    config.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    config.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    config.declare(
        "num_parallel_inlet_stages",
        ConfigValue(
            default=4,
            domain=int,
            description="Number of parallel inlet stages to simulate partial arc "
                        "admission.  Default=4",
        ),
    )
    config.declare(
        "throttle_valve_function",
        ConfigValue(
            default=ValveFunctionType.linear,
            domain=In(ValveFunctionType),
            description="Valve function type, if custom provide an expression rule",
            doc="""The type of valve function, if custom provide an expression rule
with the valve_function_rule argument.
**default** - ValveFunctionType.linear
**Valid values** - {
ValveFunctionType.linear,
ValveFunctionType.quick_opening,
ValveFunctionType.equal_percentage,
ValveFunctionType.custom}""",
        ),
    )
    config.declare(
        "throttle_valve_function_callback",
        ConfigValue(
            default=None,
            description="A callback to add a custom valve function to the "
                "throttle valves or None.  If a callback is provided, it should "
                "take the valve block data as an argument and add a "
                "valve_function expressions to it. Default=None",
        ),
    )
    config.declare(
        "num_hp",
        ConfigValue(
            default=2,
            domain=int,
            description="Number of high pressure stages not including inlet stage",
            doc="Number of high pressure stages not including inlet stage",
        ),
    )
    config.declare(
        "num_ip",
        ConfigValue(
            default=10,
            domain=int,
            description="Number of intermediate pressure stages",
            doc="Number of intermediate pressure stages",
        ),
    )
    config.declare(
        "num_lp",
        ConfigValue(
            default=5,
            domain=int,
            description="Number of low pressure stages not including outlet stage",
            doc="Number of low pressure stages not including outlet stage",
        ),
    )
    config.declare(
        "hp_split_locations",
        ConfigList(
            default=[],
            domain=int,
            description="Locations of splitters in HP section",
            doc="A list of index locations of splitters in the HP section. The "
                "indexes indicate after which stage to include splitters.  0 is "
                "between the inlet stage and the first regular HP stage.",
        ),
    )
    config.declare(
        "ip_split_locations",
        ConfigList(
            default=[],
            domain=int,
            description="Locations of splitters in IP section",
            doc="A list of index locations of splitters in the IP section. The "
                "indexes indicate after which stage to include splitters.",
        ),
    )
    config.declare(
        "lp_split_locations",
        ConfigList(
            default=[],
            domain=int,
            description="Locations of splitter in LP section",
            doc="A list of index locations of splitters in the LP section. The "
                "indexes indicate after which stage to include splitters.",
        ),
    )
    config.declare(
        "hp_disconnect",
        ConfigList(
            default=[],
            domain=int,
            description="HP Turbine stages to not connect to next with an arc.",
            doc="HP Turbine stages to not connect to next with an arc. This is "
                "usually used to insert additional units between stages on a "
                "flowsheet, such as a reheater",
        ),
    )
    config.declare(
        "ip_disconnect",
        ConfigList(
            default=[],
            domain=int,
            description="IP Turbine stages to not connect to next with an arc.",
            doc="IP Turbine stages to not connect to next with an arc. This is "
            "usually used to insert additional units between stages on a "
            "flowsheet, such as a reheater",
        ),
    )
    config.declare(
        "lp_disconnect",
        ConfigList(
            default=[],
            domain=int,
            description="LP Turbine stages to not connect to next with an arc.",
            doc="LP Turbine stages to not connect to next with an arc. This is "
            "usually used to insert additional units between stages on a "
            "flowsheet, such as a reheater",
        ),
    )
    config.declare(
        "hp_split_num_outlets",
        ConfigValue(
            default={},
            domain=dict,
            description="Dict, hp split index: number of splitter outlets, if not 2",
        ),
    )
    config.declare(
        "ip_split_num_outlets",
        ConfigValue(
            default={},
            domain=dict,
            description="Dict, ip split index: number of splitter outlets, if not 2",
        ),
    )
    config.declare(
        "lp_split_num_outlets",
        ConfigValue(
            default={},
            domain=dict,
            description="Dict, lp split index: number of splitter outlets, if not 2",
        ),
    )


@declare_process_block_class(
    "HelmTurbineMultistage",
    doc="Multistage steam turbine with optional reheat and extraction",
)
class HelmTurbineMultistageData(UnitModelBlockData):
    CONFIG = ConfigBlock()
    _define_turbine_multistage_config(CONFIG)

    def build(self):
        super().build()
        config = self.config
        unit_cfg = {  # general unit model config
            "dynamic": config.dynamic,
            "has_holdup": config.has_holdup,
            "property_package": config.property_package,
            "property_package_args": config.property_package_args,
        }
        ni = self.config.num_parallel_inlet_stages
        inlet_idx = self.inlet_stage_idx = pyo.RangeSet(ni)

        thrtl_cfg = unit_cfg.copy()
        thrtl_cfg["valve_function"] = self.config.throttle_valve_function
        thrtl_cfg["valve_function_callback"] = \
            self.config.throttle_valve_function_callback

        # Adding unit models
        # ------------------------

        # Splitter to inlet that splits main flow into parallel flows for
        # paritial arc admission to the turbine
        self.inlet_split = HelmSplitter(default=self._split_cfg(unit_cfg, ni))
        self.throttle_valve = SteamValve(inlet_idx, default=thrtl_cfg)
        self.inlet_stage = HelmTurbineInletStage(inlet_idx, default=unit_cfg)
        # mixer to combine the parallel flows back together
        self.inlet_mix = HelmMixer(default=self._mix_cfg(unit_cfg, ni))
        # add turbine sections.
        # inlet stage -> hp stages -> ip stages -> lp stages -> outlet stage
        self.hp_stages = HelmTurbineStage(pyo.RangeSet(config.num_hp), default=unit_cfg)
        self.ip_stages = HelmTurbineStage(pyo.RangeSet(config.num_ip), default=unit_cfg)
        self.lp_stages = HelmTurbineStage(pyo.RangeSet(config.num_lp), default=unit_cfg)
        self.outlet_stage = HelmTurbineOutletStage(default=unit_cfg)

        for i in self.hp_stages:
            self.hp_stages[i].ratioP.fix()
            self.hp_stages[i].efficiency_isentropic.fix()
        for i in self.ip_stages:
            self.ip_stages[i].ratioP.fix()
            self.ip_stages[i].efficiency_isentropic.fix()
        for i in self.lp_stages:
            self.lp_stages[i].ratioP.fix()
            self.lp_stages[i].efficiency_isentropic.fix()

        # Then make splitter config.  If number of outlets is specified
        # make a specific config, otherwise use default with 2 outlets
        s_sfg_default = self._split_cfg(unit_cfg, 2)
        hp_splt_cfg = {}
        ip_splt_cfg = {}
        lp_splt_cfg = {}
        # Now to finish up if there are more than two outlets, set that
        for i, v in config.hp_split_num_outlets.items():
            hp_splt_cfg[i] = self._split_cfg(unit_cfg, v)
        for i, v in config.ip_split_num_outlets.items():
            ip_splt_cfg[i] = self._split_cfg(unit_cfg, v)
        for i, v in config.lp_split_num_outlets.items():
            lp_splt_cfg[i] = self._split_cfg(unit_cfg, v)
        # put in splitters for turbine steam extractions
        if config.hp_split_locations:
            self.hp_split = HelmSplitter(
                config.hp_split_locations,
                default=s_sfg_default,
                initialize=hp_splt_cfg
            )
        else:
            self.hp_split = {}
        if config.ip_split_locations:
            self.ip_split = HelmSplitter(
                config.ip_split_locations,
                default=s_sfg_default,
                initialize=ip_splt_cfg
            )
        else:
            self.ip_split = {}
        if config.lp_split_locations:
            self.lp_split = HelmSplitter(
                config.lp_split_locations,
                default=s_sfg_default,
                initialize=lp_splt_cfg
            )
        else:
            self.lp_split = {}
        # Done with unit models.  Adding Arcs (streams).
        # ------------------------------------------------

        # First up add streams in the inlet section
        def _split_to_rule(b, i):
            return {
                "source": getattr(self.inlet_split, "outlet_{}".format(i)),
                "destination": self.throttle_valve[i].inlet,
            }
        def _valve_to_rule(b, i):
            return {
                "source": self.throttle_valve[i].outlet,
                "destination": self.inlet_stage[i].inlet,
            }
        def _inlet_to_rule(b, i):
            return {
                "source": self.inlet_stage[i].outlet,
                "destination": getattr(self.inlet_mix, "inlet_{}".format(i)),
            }
        self.stream_throttle_inlet = Arc(inlet_idx, rule=_split_to_rule)
        self.stream_throttle_outlet = Arc(inlet_idx, rule=_valve_to_rule)
        self.stream_inlet_mix_inlet = Arc(inlet_idx, rule=_inlet_to_rule)

        # There are three sections HP, IP, and LP which all have the same sort
        # of internal connctions, so the functions below provide some generic
        # capcbilities for adding the internal Arcs (streams).
        def _arc_indexes(nstages, index_set, discon, splits):
            """
            This takes the index set of all possible streams in a turbine
            section and throws out arc indexes for stages that are disconnected
            and arc indexes that are not needed because there is no splitter
            after a stage.

            Args:
                nstages (int): Number of stages in section
                index_set (Set): Index set for arcs in the section
                discon (list): Disconnected stages in the section
                splits (list): Spliter locations
            """
            sr = set()  # set of things to remove from the Arc index set
            for i in index_set:
                if (i[0] in discon or i[0] == nstages) and i[0] in splits:
                    # don't connect stage i to next remove stream after split
                    sr.add((i[0], 2))
                elif (i[0] in discon or i[0] == nstages) and i[0] not in splits:
                    # no splitter and disconnect so remove both streams
                    sr.add((i[0], 1))
                    sr.add((i[0], 2))
                elif i[0] not in splits:
                    # no splitter and not disconnected so just second stream
                    sr.add((i[0], 2))
                else:
                    # has splitter so need both streams don't remove anything
                    pass
            for i in sr:  # remove the unneeded Arc indexes
                index_set.remove(i)

        def _arc_rule(turbines, splitters):
            """
            This creates a rule function for arcs in a turbine section. When
            this is used, the indexes for nonexistant stream will have already
            been removed, so any indexes the rule will get should have a stream
            associated.

            Args:
                turbines (TurbineStage): Indexed block with turbine section stages
                splitters (Separator): Indexed block of splitters
            """

            def _rule(b, i, j):
                if i in splitters and j == 1: # stage to splitter
                    return {
                        "source": turbines[i].outlet,
                        "destination": splitters[i].inlet,
                    }
                elif j == 2: # splitter to next stage
                    return {
                        "source": splitters[i].outlet_1,
                        "destination": turbines[i + 1].inlet,
                    }
                else: # no splitter, stage to next stage
                    return {
                        "source": turbines[i].outlet,
                        "destination": turbines[i + 1].inlet,
                    }

            return _rule

        # Create initial arcs index sets with all possible streams
        self.hp_stream_idx = pyo.Set(initialize=self.hp_stages.index_set() * [1, 2])
        self.ip_stream_idx = pyo.Set(initialize=self.ip_stages.index_set() * [1, 2])
        self.lp_stream_idx = pyo.Set(initialize=self.lp_stages.index_set() * [1, 2])

        # Throw out unneeded streams for disconnected stages or no splitter
        _arc_indexes(
            config.num_hp,
            self.hp_stream_idx,
            config.hp_disconnect,
            config.hp_split_locations,
        )
        _arc_indexes(
            config.num_ip,
            self.ip_stream_idx,
            config.ip_disconnect,
            config.ip_split_locations,
        )
        _arc_indexes(
            config.num_lp,
            self.lp_stream_idx,
            config.lp_disconnect,
            config.lp_split_locations,
        )

        # Create connections internal to each turbine section (hp, ip, and lp)
        self.hp_stream = Arc(
            self.hp_stream_idx, rule=_arc_rule(self.hp_stages, self.hp_split))
        self.ip_stream = Arc(
            self.ip_stream_idx, rule=_arc_rule(self.ip_stages, self.ip_split))
        self.lp_stream = Arc(
            self.lp_stream_idx, rule=_arc_rule(self.lp_stages, self.lp_split))

        # Connect hp section to ip section unless its a disconnect location
        last_hp = config.num_hp
        if 0 not in config.ip_disconnect and last_hp not in config.hp_disconnect:
            # Not disconnected stage so add stream, depending on splitter existance
            if last_hp in config.hp_split_locations:  # connect splitter to ip
                self.hp_to_ip_stream = Arc(
                    source=self.hp_split[last_hp].outlet_1,
                    destination=self.ip_stages[1].inlet,
                )
            else:  # connect last hp to ip
                self.hp_to_ip_stream = Arc(
                    source=self.hp_stages[last_hp].outlet,
                    destination=self.ip_stages[1].inlet,
                )
        # Connect ip section to lp section unless its a disconnect location
        last_ip = config.num_ip
        if 0 not in config.lp_disconnect and last_ip not in config.ip_disconnect:
            if last_ip in config.ip_split_locations:  # connect splitter to ip
                self.ip_to_lp_stream = Arc(
                    source=self.ip_split[last_ip].outlet_1,
                    destination=self.lp_stages[1].inlet,
                )
            else:  # connect last hp to ip
                self.ip_to_lp_stream = Arc(
                    source=self.ip_stages[last_ip].outlet,
                    destination=self.lp_stages[1].inlet,
                )
        # Connect inlet stage to hp section
        #   not allowing disconnection of inlet and first regular hp stage
        if 0 in config.hp_split_locations:
            # connect inlet mix to splitter and splitter to hp section
            self.inlet_to_splitter_stream = Arc(
                source=self.inlet_mix.outlet, destination=self.hp_split[0].inlet
            )
            self.splitter_to_hp_stream = Arc(
                source=self.hp_split[0].outlet_1, destination=self.hp_stages[1].inlet
            )
        else:  # connect mixer to first hp turbine stage
            self.inlet_to_hp_stream = Arc(
                source=self.inlet_mix.outlet, destination=self.hp_stages[1].inlet
            )

        self.power = pyo.Var(
            self.flowsheet().time, initialize=-1e8, doc="power (W)")
        @self.Constraint(self.flowsheet().time)
        def power_eqn(b, t):
            return (b.power[t] ==
                b.outlet_stage.control_volume.work[t]*b.outlet_stage.efficiency_mech
                + sum(
                    b.inlet_stage[i].control_volume.work[t]*b.inlet_stage[i].efficiency_mech
                    for i in b.inlet_stage)
                + sum(
                    b.hp_stages[i].control_volume.work[t]*b.hp_stages[i].efficiency_mech
                    for i in b.hp_stages)
                + sum(
                    b.ip_stages[i].control_volume.work[t]*b.ip_stages[i].efficiency_mech
                    for i in b.ip_stages)
                + sum(
                    b.lp_stages[i].control_volume.work[t]*b.lp_stages[i].efficiency_mech
                    for i in b.lp_stages)
            )


        # Connect lp section to outlet stage, not allowing outlet stage to be
        # disconnected
        last_lp = config.num_lp
        if last_lp in config.lp_split_locations:  # connect splitter to outlet
            self.lp_to_outlet_stream = Arc(
                source=self.lp_split[last_lp].outlet_1,
                destination=self.outlet_stage.inlet,
            )
        else:  # connect last lpstage to outlet
            self.lp_to_outlet_stream = Arc(
                source=self.lp_stages[last_lp].outlet,
                destination=self.outlet_stage.inlet,
            )
        pyo.TransformationFactory("network.expand_arcs").apply_to(self)

    def _split_cfg(self, unit_cfg, no=2):
        """
        This creates a configuration dictionary for a splitter.

        Args:
            unit_cfg: The base unit config dict.
            no: Number of outlets, default=2
        """
        # Create a dict for splitter config args
        cfg = copy.copy(unit_cfg)
        cfg.update(num_outlets=no)
        return cfg

    def _mix_cfg(self, unit_cfg, ni=2):
        """
        This creates a configuration dictionary for a mixer.

        Args:
            unit_cfg: The base unit config dict.
            ni: Number of inlets, default=2
        """
        cfg = copy.copy(unit_cfg)
        cfg.update(
            num_inlets=ni,
            momentum_mixing_type=MomentumMixingType.minimize_and_equality
        )
        return cfg

    def throttle_cv_fix(self, value):
        """
        Fix the thottle valve coefficients.  These are generally the same for
        each of the parallel stages so this provides a convenient way to set
        them.

        Args:
            value: The value to fix the turbine inlet flow coefficients at
        """
        for i in self.throttle_valve:
            self.throttle_valve[i].Cv.fix(value)

    def turbine_inlet_cf_fix(self, value):
        """
        Fix the inlet turbine stage flow coefficient.  These are
        generally the same for each of the parallel stages so this provides
        a convenient way to set them.

        Args:
            value: The value to fix the turbine inlet flow coefficients at
        """
        for i in self.inlet_stage:
            self.inlet_stage[i].flow_coeff.fix(value)


    def _init_section(
        self,
        stages,
        splits,
        disconnects,
        prev_port,
        outlvl,
        solver,
        optarg,
        copy_disconneted_flow,
        copy_disconneted_pressure,
    ):
        """ Reuse the initializtion for HP, IP and, LP sections.
        """
        if 0 in splits:
            copy_port(splits[0].inlet, prev_port)
            splits[0].initialize(outlvl=outlvl, solver=solver, optarg=optarg)
            prev_port = splits[0].outlet_1
        for i in stages:
            if i - 1 not in disconnects:
                copy_port(stages[i].inlet, prev_port)
            else:
                if copy_disconneted_flow:
                    for t in stages[i].inlet.flow_mol:
                        stages[i].inlet.flow_mol[t] = pyo.value(prev_port.flow_mol[t])
                if copy_disconneted_pressure:
                    for t in stages[i].inlet.pressure:
                        stages[i].inlet.pressure[t] = pyo.value(prev_port.pressure[t])
            stages[i].initialize(outlvl=outlvl, solver=solver, optarg=optarg)
            prev_port = stages[i].outlet
            if i in splits:
                copy_port(splits[i].inlet, prev_port)
                splits[i].initialize(outlvl=outlvl, solver=solver, optarg=optarg)
                prev_port = splits[i].outlet_1
        return prev_port


    def turbine_outlet_cf_fix(self, value):
        """
        Fix the inlet turbine stage flow coefficient.  These are
        generally the same for each of the parallel stages so this provides
        a convenient way to set them.

        Args:
            value: The value to fix the turbine inlet flow coefficients at
        """
        self.outlet_stage.flow_coeff.fix(value)

    def initialize(
        self,
        outlvl=idaeslog.NOTSET,
        solver=None,
        flow_iterate=2,
        optarg=None,
        copy_disconneted_flow=True,
        copy_disconneted_pressure=True,
        calculate_outlet_cf=False,
        calculate_inlet_cf=False
    ):
        """
        Initialize

        Args:
            outlvl: logging level default is NOTSET, which inherits from the
                parent logger
            solver: the NL solver
            flow_iterate: If not calculating flow coefficients, this is the
                number of times to update the flow and repeat initialization
                (1 to 5 where 1 does not update the flow guess)
            optarg: solver arguments, default is None
            copy_disconneted_flow: Copy the flow through the disconnected stages
                default is True
            copy_disconneted_pressure: Copy the pressure through the disconnected
                stages default is True
            calculate_outlet_cf: Use the flow initial flow guess to calculate
                the outlet stage flow coefficient, default is False,
            calculate_inlet_cf: Use the inlet stage ratioP to calculate the flow
                coefficent for the inlet stage default is False

        Returns:
            None
        """
        # Setup loggers
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        # Store initial model specs, restored at the end of initializtion, so
        # the problem is not altered.  This can restore fixed/free vars,
        # active/inactive constraints, and fixed variable values.
        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)

        # Assume the flow into the turbine is a reasonable guess for
        # initializtion
        flow_guess = self.inlet_split.inlet.flow_mol[0].value

        for it_count in range(flow_iterate):
            self.inlet_split.initialize(
                outlvl=outlvl, solver=solver, optarg=optarg)

            # Initialize valves
            for i in self.inlet_stage_idx:
                u = self.throttle_valve[i]
                copy_port(u.inlet, getattr(
                    self.inlet_split, "outlet_{}".format(i)))
                u.initialize(
                    outlvl=outlvl, solver=solver, optarg=optarg)

            # Initialize turbine
            for i in self.inlet_stage_idx:
                u = self.inlet_stage[i]
                copy_port(u.inlet, self.throttle_valve[i].outlet)
                u.initialize(
                    outlvl=outlvl,
                    solver=solver,
                    optarg=optarg,
                    calculate_cf=calculate_inlet_cf
                )

            # Initialize Mixer
            self.inlet_mix.use_minimum_inlet_pressure_constraint()
            for i in self.inlet_stage_idx:
                copy_port(
                    getattr(self.inlet_mix, "inlet_{}".format(i)),
                    self.inlet_stage[i].outlet,
                )
                getattr(self.inlet_mix, "inlet_{}".format(i)).fix()
            self.inlet_mix.initialize(
                outlvl=outlvl, solver=solver, optarg=optarg)
            for i in self.inlet_stage_idx:
                getattr(self.inlet_mix, "inlet_{}".format(i)).unfix()
            self.inlet_mix.use_equal_pressure_constraint()

            prev_port = self.inlet_mix.outlet
            prev_port = self._init_section(
                self.hp_stages,
                self.hp_split,
                self.config.hp_disconnect,
                prev_port,
                outlvl,
                solver,
                optarg,
                copy_disconneted_flow=copy_disconneted_flow,
                copy_disconneted_pressure=copy_disconneted_pressure,
            )
            if len(self.hp_stages) in self.config.hp_disconnect:
                self.config.ip_disconnect.append(0)
            prev_port = self._init_section(
                self.ip_stages,
                self.ip_split,
                self.config.ip_disconnect,
                prev_port,
                outlvl,
                solver,
                optarg,
                copy_disconneted_flow=copy_disconneted_flow,
                copy_disconneted_pressure=copy_disconneted_pressure,
            )
            if len(self.ip_stages) in self.config.ip_disconnect:
                self.config.lp_disconnect.append(0)
            prev_port = self._init_section(
                self.lp_stages,
                self.lp_split,
                self.config.lp_disconnect,
                prev_port,
                outlvl,
                solver,
                optarg,
                copy_disconneted_flow=copy_disconneted_flow,
                copy_disconneted_pressure=copy_disconneted_pressure,
            )

            copy_port(self.outlet_stage.inlet, prev_port)
            self.outlet_stage.initialize(
                outlvl=outlvl,
                solver=solver,
                optarg=optarg,
                calculate_cf=calculate_outlet_cf
            )
            if calculate_outlet_cf:
                break
            if it_count < flow_iterate - 1:
                for t in self.inlet_split.inlet.flow_mol:
                    self.inlet_split.inlet.flow_mol[t].value = \
                        self.outlet_stage.inlet.flow_mol[t].value

                    for s in self.hp_split.values():
                        for i, o in enumerate(s.outlet_list):
                            if i == 0:
                                continue
                            o = getattr(s, o)
                            self.inlet_split.inlet.flow_mol[t].value += \
                                o.flow_mol[t].value
                    for s in self.ip_split.values():
                        for i, o in enumerate(s.outlet_list):
                            if i == 0:
                                continue
                            o = getattr(s, o)
                            self.inlet_split.inlet.flow_mol[t].value += \
                                o.flow_mol[t].value
                    for s in self.lp_split.values():
                        for i, o in enumerate(s.outlet_list):
                            if i == 0:
                                continue
                            o = getattr(s, o)
                            self.inlet_split.inlet.flow_mol[t].value += \
                                o.flow_mol[t].value


        if calculate_inlet_cf:
            # cf was probably fixed, so will have to set the value agian here
            # if you ask for it to be calculated.
            icf = {}
            for i in self.inlet_stage:
                for t in self.inlet_stage[i].flow_coeff:
                    icf[i,t] = pyo.value(self.inlet_stage[i].flow_coeff[t])
        if calculate_outlet_cf:
            ocf = pyo.value(self.outlet_stage.flow_coeff)

        from_json(self, sd=istate, wts=sp)

        if calculate_inlet_cf:
            # cf was probably fixed, so will have to set the value agian here
            # if you ask for it to be calculated.
            for t in self.inlet_stage[i].flow_coeff:
                for i in self.inlet_stage:
                    self.inlet_stage[i].flow_coeff[t] = icf[i,t]
        if calculate_outlet_cf:
            self.outlet_stage.flow_coeff = ocf


    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        # Add a default power scale
        # pretty safe to say power is around 100 to 1000 MW

        for t in self.power:
            if iscale.get_scaling_factor(self.power[t]) is None:
                iscale.set_scaling_factor(self.power[t], 1e-8)

        for t, c in self.power_eqn.items():
            power_scale = iscale.get_scaling_factor(
                self.power[t], default=1, warning=True)
            # Set power equation scale factor
            iscale.constraint_scaling_transform(
                c, power_scale, overwrite=False)
